"""
Definition for all enneoxOS senosrs
"""

from .const import Identifier, SMATagList
from .sensor import Sensor

""" The profiles are necessary because the devices generally
    return a significantly lower number of measured values without sunlight. """
ennexosSensorProfiles: list[tuple[list[int], list[str]]] = [
    (
        [9492, 9491, 9489, 9488],  # Tripower X
        [
            "Coolsys.Inverter.TmpVal.1",
            "Coolsys.Inverter.TmpVal.2",
            "Coolsys.Inverter.TmpVal.3",
            "DcMs.Amp.1",
            "DcMs.Amp.2",
            "DcMs.Amp.3",
            "DcMs.TotDcEnCntWh.1",
            "DcMs.TotDcEnCntWh.2",
            "DcMs.TotDcEnCntWh.3",
            "DcMs.Vol.1",
            "DcMs.Vol.2",
            "DcMs.Vol.3",
            "DcMs.Watt.1",
            "DcMs.Watt.2",
            "DcMs.Watt.3",
            "GridGuard.Cntry",
            "GridMs.A.phsA",
            "GridMs.A.phsB",
            "GridMs.A.phsC",
            "GridMs.GriTyp",
            "GridMs.Hz",
            "GridMs.PhV.phsA",
            "GridMs.PhV.phsA2B",
            "GridMs.PhV.phsB",
            "GridMs.PhV.phsB2C",
            "GridMs.PhV.phsC",
            "GridMs.PhV.phsC2A",
            "GridMs.TotA",
            "GridMs.TotPFEEI",
            "GridMs.TotPFExt",
            "GridMs.TotPFPrc",
            "GridMs.TotVA",
            "GridMs.TotVAr",
            "GridMs.TotW",
            "GridMs.TotW.Pv",
            "GridMs.VA.phsA",
            "GridMs.VA.phsB",
            "GridMs.VA.phsC",
            "GridMs.VAr.phsA",
            "GridMs.VAr.phsB",
            "GridMs.VAr.phsC",
            "GridMs.W.phsA",
            "GridMs.W.phsB",
            "GridMs.W.phsC",
            "InOut.GI1",  # codespell:ignore
            "Inverter.VArModCfg.PFCtlVolCfg.Stt",
            "Isolation.FltA",
            "Isolation.LeakRis",
            "Metering.TotFeedTms",
            "Metering.TotOpTms",
            "Metering.TotWhOut",
            "Metering.TotWhOut.Pv",
            "Operation.BckStt",
            "Operation.DrtStt",
            "Operation.Evt.Dsc",
            "Operation.Evt.EvtNo",
            "Operation.Evt.Msg",
            "Operation.EvtCntIstl",
            "Operation.EvtCntUsr",
            "Operation.GriSwCnt",
            "Operation.GriSwStt",
            "Operation.Health",
            "Operation.HealthStt.Alm",
            "Operation.HealthStt.Ok",
            "Operation.HealthStt.Wrn",
            "Operation.OpStt",
            "Operation.PvGriConn",
            "Operation.RstrLokStt",
            "Operation.RunStt",
            "Operation.StandbyStt",
            "Operation.VArCtl.VArModAct",
            "Operation.VArCtl.VArModStt",
            "Operation.WMaxLimSrc",
            "Operation.WMinLimSrc",
            "PvGen.PvW",
            "PvGen.PvWh",
            "Spdwr.ComSocA.Stt",
            "SunSpecSig.SunSpecTx.1",
            "Upd.Stt",
            "WebConn.Stt",
            "Wl.AcqStt",
            "Wl.AntMod",
            "Wl.ConnStt",
            "Wl.SigPwr",
            "Wl.SoftAcsConnStt",
            "Setpoint.PlantControl.InOut.GO1",  # codespell:ignore
            "Inverter.WMax",
        ],
    ),
    # Charger
    (
        [9484],
        [
            "ChaSess.WhIn",
            "Chrg.ModSw",
            "GridMs.A.phsA",
            "GridMs.A.phsB",
            "GridMs.A.phsC",
            "GridMs.Hz",
            "GridMs.PhV.phsA",
            "GridMs.PhV.phsB",
            "GridMs.PhV.phsC",
            "GridMs.TotPF",
            "GridMs.TotVA",
            "GridMs.TotVAr",
            "InOut.GI1",  # codespell:ignore
            "Metering.GridMs.TotWIn",
            "Metering.GridMs.TotWIn.ChaSta",
            "Metering.GridMs.TotWhIn",
            "Metering.GridMs.TotWhIn.ChaSta",
            "Operation.EVeh.ChaStt",
            "Operation.EVeh.Health",
            "Operation.Evt.Msg",
            "Operation.Health",
            "Operation.WMaxLimNom",
            "Operation.WMaxLimSrc",
            "Wl.AcqStt",
            "Wl.ConnStt",
            "Wl.SigPwr",
            "Wl.SoftAcsConnStt",
        ],
    ),
    # Sunny Boy Smart Energy 5.0
    (
        [19130],
        [
            "Coolsys.Inverter.TmpVal.1",
            "DcMs.Amp.1",
            "DcMs.Amp.2",
            "DcMs.Amp.3",
            "DcMs.TotDcEnCntWh.1",
            "DcMs.TotDcEnCntWh.2",
            "DcMs.TotDcEnCntWh.3",
            "DcMs.Vol.1",
            "DcMs.Vol.2",
            "DcMs.Vol.3",
            "DcMs.Watt.1",
            "DcMs.Watt.2",
            "DcMs.Watt.3",
            "GridGuard.Cntry",
            "GridMs.A.phsA",
            "GridMs.A.phsA",
            "GridMs.GriTyp",
            "GridMs.Hz",
            "GridMs.PhV.phsA2B",
            "GridMs.PhV.phsA",
            "GridMs.TotA",
            "GridMs.TotPFEEI",
            "GridMs.TotPFExt",
            "GridMs.TotPFPrc",
            "GridMs.TotVA",
            "GridMs.TotVAr",
            "GridMs.TotW",
            "GridMs.TotW.Pv",
            "GridMs.VA.phsA",
            "GridMs.VAr.phsA",
            "GridMs.W.phsA",
            "InOut.GI1",  # codespell:ignore
            "Inverter.VArModCfg.PFCtlVolCfg.Stt",
            "Isolation.FltA",
            "Isolation.LeakRis",
            "Metering.GridMs.TotWhIn.ChaSta",
            "Metering.GridMs.TotWhIn",
            "Metering.GridMs.TotWIn.ChaSta",
            "Metering.GridMs.TotWIn",
            "Metering.TotFeedTms",
            "Metering.TotOpTms",
            "Metering.TotWhOut",
            "Metering.TotWhOut.Pv",
            "Operation.BckStt",
            "Operation.DrtStt",
            "Operation.EvtCntIstl",
            "Operation.EvtCntUsr",
            "Operation.Evt.Dsc",
            "Operation.Evt.Msg",
            "Operation.GriSwCnt",
            "Operation.GriSwStt",
            "Operation.Health",
            "Operation.HealthStt.Alm",
            "Operation.HealthStt.Ok",
            "Operation.HealthStt.Wrn",
            "Operation.OpStt",
            "Operation.PvGriConn",
            "Operation.RstrLokStt",
            "Operation.RunStt",
            "Operation.StandbyStt",
            "Operation.VArCtl.VArModAct",
            "Operation.WMaxLimSrc",
            "PvGen.PvWh",
            "PvGen.PvW",
            "Setpoint.PlantControl.InOut.GO1",  # codespell:ignore
            "Spdwr.ComSocA.Stt",
            "SunSpecSig.SunSpecTx.1",
            "Upd.Stt",
            "WebConn.Stt",
            "Wl.AcqStt",
            "Wl.AntMod",
            "Wl.ConnStt",
            "Wl.SigPwr",
            "Wl.SoftAcsConnStt",
            "Bat.Amp.1",
            "Bat.ChaSttArr.1",
            "Bat.Diag.VolMax.1",
            "Bat.OpTms.1",
            "Bat.Vol.1",
            "BatChrg.BatChrg",
            "BatChrg.BatChrgArr.1",
            "BatChrg.CurBatChaArr.1",
            "BatDsch.BatDsch",
            "BatDsch.BatDschArr.1",
            "BatDsch.CurBatDschArr.1",
            "DcMs.TotDcEnCntWh.1",
            "DcMs.TotDcEnCntWh.2",
            "DcMs.TotDcEnCntWh.3",
            "EpsMs.Amp",
            "EpsMs.Vol",
            "EpsMs.Watt",
            "GridGuard.Cntry",
            "GridMs.GriTyp",
            "GridMs.TotPFEEI",
            "GridMs.TotPFExt",
            "GridMs.TotPFPrc",
            "InOut.GI1",  # codespell:ignore
            "InOut.GO1",  # codespell:ignore
            "Inverter.VArModCfg.PFCtlVolCfg.Stt",
            "Isolation.LeakRis",
            "Metering.GridMs.TotWhIn.Bat",
            "Metering.GridMs.TotWhOut.Bat",
            "Metering.TotFeedTms",
            "Metering.TotOpTms",
            "Metering.TotWhIn",
            "Operation.Bat.Health",
            "Operation.BckStt",
            "Operation.DrtStt",
            "Operation.Evt.Dsc",
            "Operation.Evt.Msg",
            "Operation.EvtCntIstl",
            "Operation.EvtCntUsr",
            "Operation.GriSwCnt",
            "Operation.HealthStt.Alm",
            "Operation.HealthStt.Ok",
            "Operation.HealthStt.Wrn",
            "Operation.PvGriConn",
            "Operation.RstrLokStt",
            "Operation.RunStt",
            "Operation.StandbyStt",
            "Operation.VArCtl.VArModAct",
            "Operation.VArCtl.VArModStt",
            "Operation.WMaxLimSrc",
            "Operation.WMinLimSrc",
            "Portal.EnnexOS.ConnStt",
            "PvGen.PvW",
            "PvGen.PvWh",
            "Spdwr.ComSocA.Stt",
            "SunSpecSig.SunSpecTx.1",
            "Upd.Stt",
            "WebConn.Stt",
            "Wl.AcqStt",
            "Wl.AntMod",
            "Wl.ConnStt",
            "Wl.SigPwr",
            "Wl.SoftAcsConnStt",
            "Setpoint.PlantControl.CmpBMS.OpMod.1",
            "Metering.GridMs.TotWhIn.Bat",
            "Metering.GridMs.TotWhOut.Bat",
            "Inverter.WMax",
        ],
    ),
    # Data Manager EDMM-10
    (
        [9397],
        [
            "Operation.Health",
            "DevUpd.IsOn",
            "FTPPush.IsOn",
            "FTPPush.Login",
            "FTPPush.Port",
            "FTPPush.Pwd",
            "FTPPush.RemStoLcn",
            "FTPPush.Srv",
            "FTPPush.UpldCyc",
            "InOut.GI1",  # codespell:ignore
            "Nameplate.CmpMain.HwRev",
            "Nameplate.CmpMain.Rev",
            "Nameplate.CmpOS.SwRev",
            "Nameplate.Location",
            "Nameplate.MainModel",
            "Nameplate.Model",
            "Nameplate.ModelStr",
            "Nameplate.PkgRev",
            "Nameplate.SerNum",
            "Nameplate.Vendor",
            "Portal.EnnexOS.ConnStt",
            "Spdwr.ActlDnsSrvIp",
            "Spdwr.ActlGwIp",
            "Spdwr.ActlIp",
            "Spdwr.ActlSnetMsk",
            "Spdwr.AutoCfgIsOn",
            "Spdwr.IgmpQryTms",
            "Spdwr.IgmpQryTx",
            "SwCmp.CmpEnnexOS.Frwk.SwRev",
            "Sys.DevRstr",
            "Upd.AutoUpdIsOn",
            "WebConn.IsOn",
            "WebConn.Stt",
            "WebConn.SwRev",
        ],
    ),
    # Energy Meter
    (
        [9307],
        [
            "Metering.GridMs.TotWIn",
            "Metering.GridMs.TotWhIn",
            "Metering.GridMs.TotWOut",
            "Metering.GridMs.TotWhOut",
            "Metering.GridMs.A.phsA",
            "Metering.GridMs.A.phsB",
            "Metering.GridMs.A.phsC",
            "Metering.GridMs.Hz",
            "Metering.GridMs.PhV.phsA",
            "Metering.GridMs.PhV.phsB",
            "Metering.GridMs.PhV.phsC",
            "Metering.GridMs.TotPF",
            "Metering.GridMs.W.phsA",
            "Metering.GridMs.W.phsB",
            "Metering.GridMs.W.phsC",
            "Operation.Health",
            "Metering.GridMs.VA.phsA",
            "Metering.GridMs.VA.phsB",
            "Metering.GridMs.VA.phsC",
            "Metering.GridMs.VAr.phsA",
            "Metering.GridMs.VAr.phsB",
            "Metering.GridMs.VAr.phsC",
        ],
    ),
    # Fake Number for Plants
    (
        [-47114711],
        [
            "GridGuard.Cntry.FrqCtl.ReconMax",
            "GridGuard.Cntry.FrqCtl.ReconMin",
            "GridGuard.Cntry.GriFltMonTms",
            "GridGuard.Cntry.GriFltReConTms",
            "GridGuard.Cntry.GriFltTms",
            "GridGuard.Cntry.GriStrTms",
            "GridGuard.Cntry.VolCtl.ReconMaxPu",
            "GridGuard.Cntry.VolCtl.ReconMinPu",
            "GridMs.TotVAr",
            "GridMs.TotW",
            "GridMs.TotW.Pv",
            "Inverter.CtlComCfg.PFCtlCom.CtlComMssMod",
            "Inverter.CtlComCfg.PFCtlCom.FlbPFExtIn",
            "Inverter.CtlComCfg.PFCtlCom.FlbPFExtOut",
            "Inverter.CtlComCfg.PFCtlCom.FlbPFIn",
            "Inverter.CtlComCfg.PFCtlCom.FlbPFOut",
            "Inverter.CtlComCfg.PFCtlCom.TmsOut",
            "Inverter.CtlComCfg.PFCtlEvu.AnInEna",
            "Inverter.CtlComCfg.PFCtlEvu.MbEna",
            "Inverter.CtlComCfg.VArCtlCom.CtlComMssMod",
            "Inverter.CtlComCfg.VArCtlCom.FlbVArNom",
            "Inverter.CtlComCfg.VArCtlCom.TmsOut",
            "Inverter.CtlComCfg.VArCtlEvu.AnInEna",
            "Inverter.CtlComCfg.VArCtlEvu.DigInEna",
            "Inverter.CtlComCfg.VArCtlEvu.MbEna",
            "Inverter.CtlComCfg.VArCtlVolCom.CtlComMssMod",
            "Inverter.CtlComCfg.VArCtlVolCom.FlbVolRefPu",
            "Inverter.CtlComCfg.VArCtlVolCom.TmsOut",
            "Inverter.CtlComCfg.WCtlCom2.CtlComMssMod",
            "Inverter.CtlComCfg.WCtlCom2.FlbWMaxNom",
            "Inverter.CtlComCfg.WCtlCom2.FlbWMinNom",
            "Inverter.CtlComCfg.WCtlCom2.TmsOut",
            "Inverter.CtlComCfg.WCtlCom.CtlComMssMod",
            "Inverter.CtlComCfg.WCtlCom.FlbWMaxNom",
            "Inverter.CtlComCfg.WCtlCom.FlbWMinNom",
            "Inverter.CtlComCfg.WCtlCom.TmsOut",
            "Inverter.CtlComCfg.WCtlDctMrk.MbEna",
            "Inverter.CtlComCfg.WCtlDctMrk.SmaSpotEna",
            "Inverter.CtlComCfg.WCtlEvu.AnInEna",
            "Inverter.CtlComCfg.WCtlEvu.DigInEna",
            "Inverter.CtlComCfg.WCtlEvu.MbEna",
            "Inverter.CurWCtlDrkNom",
            "Inverter.CurWCtlNom",
            "Inverter.DGSModCfg.AvLtTms",
            "Inverter.DGSModCfg.AvStTms",
            "Inverter.DGSModCfg.DGSFlCfg.DbVMaxPu",
            "Inverter.DGSModCfg.DGSFlCfg.DbVMinPu",
            "Inverter.DGSModCfg.DGSFlCfg.FltEnd.ArMaxTms",
            "Inverter.DGSModCfg.DGSFlCfg.StepVolPu",
            "Inverter.DGSModCfg.DGSMod",
            "Inverter.DGSModCfg.HystVolPu",
            "Inverter.DGSModCfg.VRefThMod",
            "Inverter.PFMinQ1",
            "Inverter.PFMinQ1Rtg",
            "Inverter.PFMinQ2",
            "Inverter.PFMinQ2Rtg",
            "Inverter.PFMinQ3",
            "Inverter.PFMinQ3Rtg",
            "Inverter.PFMinQ4",
            "Inverter.PFMinQ4Rtg",
            "Inverter.PlntCtl.VRef",
            "Inverter.PlntCtl.VRefMod",
            "Inverter.VALimEna",
            "Inverter.VAMaxIn",
            "Inverter.VAMaxInRtg",
            "Inverter.VAMaxOut",
            "Inverter.VAMaxOutRtg",
            "Inverter.VArGraConn",
            "Inverter.VArGraConnEna",
            "Inverter.VArMaxQ1",
            "Inverter.VArMaxQ1Rtg",
            "Inverter.VArMaxQ2",
            "Inverter.VArMaxQ2Rtg",
            "Inverter.VArMaxQ3",
            "Inverter.VArMaxQ3Rtg",
            "Inverter.VArMaxQ4",
            "Inverter.VArMaxQ4Rtg",
            "Inverter.VArMaxZerWQ1",
            "Inverter.VArMaxZerWQ2",
            "Inverter.VArMaxZerWQ3",
            "Inverter.VArMaxZerWQ4",
            "Inverter.VArModCfg.HystTms",
            "Inverter.VArModCfg.InWNomLimAct",
            "Inverter.VArModCfg.InWNomLimDeAct",
            "Inverter.VArModCfg.OutWNomLimAct",
            "Inverter.VArModCfg.OutWNomLimDeAct",
            "Inverter.VArModCfg.PFCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.PFCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.PFCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.PFCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.PFCfg.Dyn.VArTms",
            "Inverter.VArModCfg.PFCfg.Dyn.WFilTmEna",
            "Inverter.VArModCfg.PFCfg.Dyn.WFilTms",
            "Inverter.VArModCfg.PFCnstCfg.PFExtIn",
            "Inverter.VArModCfg.PFCnstCfg.PFExtOut",
            "Inverter.VArModCfg.PFCnstCfg.PFIn",
            "Inverter.VArModCfg.PFCnstCfg.PFOut",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.NumPt",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PF.1",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PF.2",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PF.3",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PF.4",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PFExt.1",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PFExt.2",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PFExt.3",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.PFExt.4",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.VolPu.1",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.VolPu.2",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.VolPu.3",
            "Inverter.VArModCfg.PFCtlVolCfg.Crv.VolPu.4",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.ActTms",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.VArTms",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.WFilTmEna",
            "Inverter.VArModCfg.PFCtlVolCfg.Dyn.WFilTms",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.NumPt",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PF.1",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PF.2",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PF.3",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PF.4",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PFExt.1",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PFExt.2",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PFExt.3",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.PFExt.4",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.WNom.1",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.WNom.2",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.WNom.3",
            "Inverter.VArModCfg.PFCtlWCfg.Crv.WNom.4",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.ActTms",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.VArTms",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.WFilTmEna",
            "Inverter.VArModCfg.PFCtlWCfg.Dyn.WFilTms",
            "Inverter.VArModCfg.PFCtlWCfg.Trg.DsaUnVolPu",
            "Inverter.VArModCfg.PFCtlWCfg.Trg.DsaVolPu",
            "Inverter.VArModCfg.PFCtlWCfg.Trg.EnaUnVolPu",
            "Inverter.VArModCfg.PFCtlWCfg.Trg.EnaVolPu",
            "Inverter.VArModCfg.PFMinEna",
            "Inverter.VArModCfg.VArCfg.Crv.CrvEna",
            "Inverter.VArModCfg.VArCfg.Crv.XVal.1",
            "Inverter.VArModCfg.VArCfg.Crv.XVal.2",
            "Inverter.VArModCfg.VArCfg.Crv.XVal.3",
            "Inverter.VArModCfg.VArCfg.Crv.XVal.4",
            "Inverter.VArModCfg.VArCfg.Crv.YVal.1",
            "Inverter.VArModCfg.VArCfg.Crv.YVal.2",
            "Inverter.VArModCfg.VArCfg.Crv.YVal.3",
            "Inverter.VArModCfg.VArCfg.Crv.YVal.4",
            "Inverter.VArModCfg.VArCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.VArCfg.Dyn.VArGraFlb",
            "Inverter.VArModCfg.VArCfg.Dyn.VArGraFlbEna",
            "Inverter.VArModCfg.VArCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.VArCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.VArCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.VArCfg.Dyn.VArTms",
            "Inverter.VArModCfg.VArCnstCfgDmd.VAr",
            "Inverter.VArModCfg.VArCnstCfgDmd.VArNom",
            "Inverter.VArModCfg.VArCnstCfgIn.VAr",
            "Inverter.VArModCfg.VArCnstCfgIn.VArNom",
            "Inverter.VArModCfg.VArCnstCfg.VAr",
            "Inverter.VArModCfg.VArCnstCfg.VArNom",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.PreSet",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.1",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.2",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.3",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.4",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.5",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.6",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.7",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.XVal.8",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.1",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.2",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.3",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.4",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.5",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.6",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.7",
            "Inverter.VArModCfg.VArCtlVolCfg.Crv.YVal.8",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.ActTms",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.VArCtlVolCfg.Dyn.VArTms",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.AutnAdjMod",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.AutnAdjTms",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.ClsCtl.DsaHiLimVArPu",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.ClsCtl.DsaLoLimVArPu",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.ClsCtl.EnaHiLimVArPu",
            "Inverter.VArModCfg.VArCtlVolCfg.VolRef.ClsCtl.EnaLoLimVArPu",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.NumPt",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.1",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.10",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.2",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.3",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.4",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.5",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.6",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.7",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.8",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.XVal.9",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.1",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.10",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.2",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.3",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.4",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.5",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.6",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.7",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.8",
            "Inverter.VArModCfg.VArCtlWCfg.Crv.YVal.9",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.ActTms",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.VArGraEna",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.VArGraNeg",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.VArGraPos",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.VArTmEna",
            "Inverter.VArModCfg.VArCtlWCfg.Dyn.VArTms",
            "Inverter.VArModCfg.VArCtlWCfg.Trg.DsaUnVolPu",
            "Inverter.VArModCfg.VArCtlWCfg.Trg.DsaVolPu",
            "Inverter.VArModCfg.VArCtlWCfg.Trg.EnaUnVolPu",
            "Inverter.VArModCfg.VArCtlWCfg.Trg.EnaVolPu",
            "Inverter.VArModCfg.VArModIn",
            "Inverter.VArModCfg.VArModInFlb",
            "Inverter.VArModCfg.VArModOut",
            "Inverter.VArModCfg.VArModOutFlb",
            "Inverter.VArModCfg.VArModZerW",
            "Inverter.VArModCfg.VArModZerWFlb",
            "Inverter.VArModCfg.VArNomRefMod",
            "Inverter.VArModCfg.VRefMod",
            "Inverter.WAval",
            "Inverter.WCtlHzModCfg.Ena",
            "Inverter.WCtlHzModCfg.HzFilTmEna",
            "Inverter.WCtlHzModCfg.HzFilTms",
            "Inverter.WCtlHzModCfg.RefModOv",
            "Inverter.WCtlHzModCfg.RefModUn",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HystEnaOv",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HystEnaUn",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOv.1",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOv.2",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOv.3",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOvGra.1",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOvGra.2",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOvGra.3",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzOvStop",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzStopWGra",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzStopWGraTms",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUn.1",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUn.2",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUn.3",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUnGra.1",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUnGra.2",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUnGra.3",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.HzUnStop",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.WCtlChgOvUnTmms",
            "Inverter.WCtlHzModCfg.WCtlHzCfg.WCtlTmms",
            "Inverter.WCtlHzModCfg.WGraEna",
            "Inverter.WCtlHzModCfg.WGraNeg",
            "Inverter.WCtlHzModCfg.WGraPos",
            "Inverter.WCtlHzModCfg.WTmEna",
            "Inverter.WCtlHzModCfg.WTms",
            "Inverter.WGraConn",
            "Inverter.WGraConnEna",
            "Inverter.WGraMpp",
            "Inverter.WGraRecon",
            "Inverter.WGraReconEna",
            "Inverter.WMax",
            "Inverter.WMaxIn",
            "Inverter.WMaxInRtg",
            "Inverter.WMaxOut",
            "Inverter.WMaxOutRtg",
            "Inverter.WModCfg.WCnstCfg.WNomEna",
            "Inverter.WModCfg.WCnstCfg.WNomPrc",
            "Inverter.WModCfg.WCtlComCfg2.Dyn.WGraEna",
            "Inverter.WModCfg.WCtlComCfg2.Dyn.WGraNeg",
            "Inverter.WModCfg.WCtlComCfg2.Dyn.WGraPos",
            "Inverter.WModCfg.WCtlComCfg2.Dyn.WTmEna",
            "Inverter.WModCfg.WCtlComCfg2.Dyn.WTms",
            "Inverter.WModCfg.WCtlComCfg2.Ena",
            "Inverter.WModCfg.WCtlComCfg2.LoPrioEna",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WGraEna",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WGraFlb",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WGraFlbEna",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WGraNeg",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WGraPos",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WTmEna",
            "Inverter.WModCfg.WCtlComCfg.Dyn.WTms",
            "Inverter.WModCfg.WCtlComCfg.Ena",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.DbHzOv",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.DbHzUn",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOv.1",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOv.2",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOv.3",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOvGra.1",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOvGra.2",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzOvGra.3",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUn.1",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUn.2",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUn.3",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUnGra.1",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUnGra.2",
            "Inverter.WModCfg.WCtlHzFsmCfg.Crv.HzUnGra.3",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.HzFilTmEna",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.HzFilTms",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WAvalFilTmEna",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WAvalGraEna",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WAvalGraNeg",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WAvalGraPos",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WAvalTms",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WDifFilTmEna",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WDifFilTms",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WDifGraEna",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WDifGraNeg",
            "Inverter.WModCfg.WCtlHzFsmCfg.Dyn.WDifGraPos",
            "Inverter.WModCfg.WCtlHzFsmCfg.Ena",
            "Inverter.WModCfg.WCtlHzFsmCfg.WBaseMod",
            "Inverter.WModCfg.WCtlHzFsmCfg.WDifRefMod",
            "Inverter.WModCfg.WCtlHzFsmCfg.WResNom",
            "Inverter.WModCfg.WCtlHzILFCfg.DsaWaitTms",
            "Inverter.WModCfg.WCtlHzILFCfg.Ena",
            "Inverter.WModCfg.WCtlHzILFCfg.HzFilTmEna",
            "Inverter.WModCfg.WCtlHzILFCfg.HzFilTms",
            "Inverter.WModCfg.WCtlHzILFCfg.HzOvDsa",
            "Inverter.WModCfg.WCtlHzILFCfg.HzOvEna",
            "Inverter.WModCfg.WCtlHzILFCfg.HzStopMax",
            "Inverter.WModCfg.WCtlHzILFCfg.HzStopMin",
            "Inverter.WModCfg.WCtlHzILFCfg.HzStopWGra",
            "Inverter.WModCfg.WCtlHzILFCfg.HzUnDsa",
            "Inverter.WModCfg.WCtlHzILFCfg.HzUnEna",
            "Inverter.WModCfg.WCtlHzILFCfg.MbEna",
            "Inverter.WModCfg.WCtlHzILFCfg.WGra",
            "Inverter.WModCfg.WCtlVolCfg.ActTms",
            "Inverter.WModCfg.WCtlVolCfg.Crv.NumPt",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.1",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.2",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.3",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.4",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.5",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.6",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.7",
            "Inverter.WModCfg.WCtlVolCfg.Crv.XVal.8",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.1",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.2",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.3",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.4",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.5",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.6",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.7",
            "Inverter.WModCfg.WCtlVolCfg.Crv.YVal.8",
            "Inverter.WModCfg.WCtlVolCfg.Ena",
            "Inverter.WModCfg.WCtlVolCfg.VRefMod",
            "Inverter.WModCfg.WCtlVolCfg.WGraEna",
            "Inverter.WModCfg.WCtlVolCfg.WGraNeg",
            "Inverter.WModCfg.WCtlVolCfg.WGraPos",
            "Inverter.WModCfg.WCtlVolCfg.WRefMod",
            "Inverter.WModCfg.WCtlVolCfg.WTmEna",
            "Inverter.WModCfg.WCtlVolCfg.WTms",
            "Metering.DyWhOut",
            "Metering.GridMs.TotWhIn.Bat",
            "Metering.GridMs.TotWhOut.Bat",
            "Metering.PCCMs.PlntA.phsA",
            "Metering.PCCMs.PlntA.phsB",
            "Metering.PCCMs.PlntA.phsC",
            "Metering.PCCMs.PlntCsmpW",
            "Metering.PCCMs.PlntCsmpWh",
            "Metering.PCCMs.PlntHz",
            "Metering.PCCMs.PlntPF",
            "Metering.PCCMs.PlntPhV.phsA",
            "Metering.PCCMs.PlntPhV.phsB",
            "Metering.PCCMs.PlntPhV.phsC",
            "Metering.PCCMs.PlntVAr",
            "Metering.PCCMs.PlntVAr.phsA",
            "Metering.PCCMs.PlntVAr.phsB",
            "Metering.PCCMs.PlntVAr.phsC",
            "Metering.PCCMs.PlntW",
            "Metering.PCCMs.PlntWh",
            "Metering.PCCMs.PlntW.phsA",
            "Metering.PCCMs.PlntW.phsB",
            "Metering.PCCMs.PlntW.phsC",
            "Metering.TotWhOut.Pv",
            "Operation.CurAvailPlnt",
            "Operation.CurAvailVArOvExt",
            "Operation.CurAvailVArOvExtNom",
            "Operation.CurAvailVArUnExt",
            "Operation.CurAvailVArUnExtNom",
            "Operation.Dmd.WCtl",
            "Operation.Health",
            "Operation.NbrQodStt",
            "Operation.NbrRunStt",
            "Operation.NbrStopStt",
            "Operation.NbrWaitStt",
            "Operation.OpStt",
            "Operation.PlntCtl.VArCtlMod",
            "Operation.PlntCtl.WCtlMod",
            "Operation.VArCtl.VArModAct",
            "Operation.VArCtl.VArModStt",
            "Operation.VArMaxLimNom",
            "Operation.VArMinLimNom",
            "Operation.WMaxInLimNom",
            "Operation.WMaxLimNom",
            "Operation.WMaxLimSrc",
            "Operation.WMinInLimNom",
            "Operation.WMinLimNom",
            "Operation.WMinLimSrc",
            "PCC.WMaxAsym",
            "PCC.WMaxAsymMod",
            "PCC.WMaxAsymRefMod",
            "Plnt.CapacRtgWh",
            "PlntCtl.Frs.BatChaSttMax",
            "PlntCtl.Frs.BatChaSttMin",
            "PlntCtl.Frs.BatWhFilTms",
            "PlntCtl.Frs.CrvActHystTms",
            "PlntCtl.Frs.CrvActTms",
            "PlntCtl.Frs.CrvRmpZerWTms",
            "PlntCtl.Frs.DbHzOv",
            "PlntCtl.Frs.DbHzUn",
            "PlntCtl.Frs.Ena",
            "PlntCtl.Frs.HystWhRsvNom",
            "PlntCtl.Frs.RsvRmpZerWTms",
            "PlntCtl.Frs.RsvTms",
            "PlntCtl.Frs.RvyWhRsv.Ena",
            "PlntCtl.Frs.RvyWhRsv.WaitTms",
            "PlntCtl.Frs.RvyWhRsv.WGra",
            "PlntCtl.Frs.RvyWhRsv.WGraEna",
            "PlntCtl.Frs.RvyWhRsv.WMaxNom",
            "PlntCtl.Frs.WMaxNom",
            "PlntCtl.MaxPhsWCtlEna",
            "PlntCtl.MsFil.HzFilTmEna",
            "PlntCtl.MsFil.HzFilTms",
            "PlntCtl.MsFil.VArFilTmEna",
            "PlntCtl.MsFil.VArFilTms",
            "PlntCtl.MsFil.VolFilTmEna",
            "PlntCtl.MsFil.VolFilTms",
            "PlntCtl.MsFil.WFilTmEna",
            "PlntCtl.MsFil.WFilTms",
            "PlntCtl.SingleDevSptEna",
            "PlntCtl.SmpTm",
            "PlntCtl.VArCtl.DeActWaitTms",
            "PlntCtl.VArCtl.WPuLimAct",
            "PlntCtl.VArCtl.WPuLimDeAct",
            "PlntCtl.WCtlEvuRefMod",
            "PlntCtl.WNomInvStop",
            "Plnt.DcWRtg",
            "Plnt.Frs.Aval",
            "Plnt.TotAvalCha",
            "Plnt.TotAvalDsch",
            "Plnt.WRtg",
            "PortalUsrSettings.PlntNam",
            "InOut.DigIn.CurWCtlNom",  # codespell:ignore
        ],
    ),
]

# fmt: off
# black -C -l 200 pysma/definitions_ennexos.py
enneoxSensors = [
        Sensor("Bat.Amp.1", None, factor=1, unit=None),
        Sensor("Bat.ChaSttArr.1", None, factor=1, unit=None),
        Sensor("BatChrg.BatChrgArr.1", None, factor=1, unit=None),
        Sensor("BatChrg.BatChrg", None, factor=1, unit=None),
        Sensor("BatChrg.CurBatChaArr.1", None, factor=1, unit=None),
        Sensor("Bat.Diag.VolMax.1", None, factor=1, unit=None),
        Sensor("BatDsch.BatDschArr.1", None, factor=1, unit=None),
        Sensor("BatDsch.BatDsch", None, factor=1, unit=None),
        Sensor("BatDsch.CurBatDschArr.1", None, factor=1, unit=None),
        Sensor("Bat.OpTms.1", None, factor=1, unit=None),
        Sensor("Bat.Vol.1", None, factor=1, unit=None),
        Sensor("ChaSess.WhIn", None, factor=1, unit=None),
        Sensor("Chrg.ModSw", None, factor=1, unit=None, mapper=SMATagList),
        Sensor("Coolsys.Inverter.TmpVal.1", Identifier.temp_a, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.2", Identifier.temp_b, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.3", Identifier.temp_c, factor=1, unit="°C"),
        Sensor("DcMs.Amp.1", Identifier.pv_current_a, factor=1, unit="A"),
        Sensor("DcMs.Amp.2", Identifier.pv_current_b, factor=1, unit="A"),
        Sensor("DcMs.Amp.3", Identifier.pv_current_c, factor=1, unit="A"),
        Sensor("DcMs.TotDcEnCntWh.1", None),
        Sensor("DcMs.TotDcEnCntWh.1", None, factor=1, unit=None),
        Sensor("DcMs.TotDcEnCntWh.2", None, factor=1, unit=None),
        Sensor("DcMs.TotDcEnCntWh.3", None, factor=1, unit=None),
        Sensor("DcMs.Vol.1", Identifier.pv_voltage_a, factor=1, unit="V"),
        Sensor("DcMs.Vol.2", Identifier.pv_voltage_b, factor=1, unit="V"),
        Sensor("DcMs.Vol.3", Identifier.pv_voltage_c, factor=1, unit="V"),
        Sensor("DcMs.Watt.1", Identifier.pv_power_a, factor=1, unit="W"),
        Sensor("DcMs.Watt.2", Identifier.pv_power_b, factor=1, unit="W"),
        Sensor("DcMs.Watt.3", Identifier.pv_power_c, factor=1, unit="W"),
        Sensor("DevUpd.IsOn", None, factor=1, unit=None),
        Sensor("EpsMs.Amp", None, factor=1, unit=None),
        Sensor("EpsMs.Vol", None, factor=1, unit=None),
        Sensor("EpsMs.Watt", None, factor=1, unit=None),
        Sensor("FTPPush.IsOn", None, factor=1, unit=None),
        Sensor("FTPPush.Login", None, factor=1, unit=None),
        Sensor("FTPPush.Port", None, factor=1, unit=None),
        Sensor("FTPPush.Pwd", None, factor=1, unit=None),
        Sensor("FTPPush.RemStoLcn", None, factor=1, unit=None),
        Sensor("FTPPush.Srv", None, factor=1, unit=None),
        Sensor("FTPPush.UpldCyc", None, factor=1, unit=None),
        Sensor("GridGuard.Cntry", None, factor=1, unit=None),
        Sensor("GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"),
        Sensor("GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("GridMs.GriTyp", None, factor=1, unit=None),
        Sensor("GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("GridMs.PhV.phsA2B", None, factor=1, unit=None),
        Sensor("GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsB2C", None, factor=1, unit=None),
        Sensor("GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsC2A", None, factor=1, unit=None),
        Sensor("GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("GridMs.TotA", Identifier.current_total, factor=1, unit="A"),
        Sensor("GridMs.TotPFEEI", None, factor=1, unit=None),
        Sensor("GridMs.TotPFExt", None, factor=1, unit=None),
        Sensor("GridMs.TotPF", None, factor=1, unit=None),
        Sensor("GridMs.TotPFPrc", None, factor=1, unit=None),
        Sensor("GridMs.TotVA", Identifier.grid_apparent_power, factor=1, unit="VA"),
        Sensor("GridMs.TotVAr", Identifier.grid_reactive_power, factor=1, unit="var"),
        Sensor("GridMs.TotW", Identifier.grid_power, factor=1, unit="W"),
        Sensor("GridMs.TotW.Pv", Identifier.pv_power, factor=1, unit="W"),
        Sensor("GridMs.VA.phsA", Identifier.grid_apparent_power_l1, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsB", Identifier.grid_apparent_power_l2, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsC", Identifier.grid_apparent_power_l3, factor=1, unit="VA"),
        Sensor("GridMs.VAr.phsA", Identifier.grid_reactive_power_l1, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsB", Identifier.grid_reactive_power_l2, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsC", Identifier.grid_reactive_power_l3, factor=1, unit="var"),
        Sensor("GridMs.W.phsA", Identifier.power_l1, factor=1, unit="W"),
        Sensor("GridMs.W.phsB", Identifier.power_l2, factor=1, unit="W"),
        Sensor("GridMs.W.phsC", Identifier.power_l3, factor=1, unit="W"),
        Sensor("InOut.GI1", None, factor=1, unit=None),  # codespell:ignore
        Sensor("InOut.GO1", None, factor=1, unit=None),  # codespell:ignore
        Sensor("Inverter.VArModCfg.PFCtlVolCfg.Stt", None, factor=1, unit=None),
        Sensor("Inverter.WMax", Identifier.inverter_power_limit, factor=1, unit="W"),
        Sensor("Isolation.FltA", Identifier.insulation_residual_current, factor=1000, unit="mA"),
        Sensor("Isolation.LeakRis", None, factor=1, unit="kOhm"),
        Sensor("Isolation.LeakRis", None, factor=1, unit=None),
        Sensor("Metering.GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"),
        Sensor("Metering.GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("Metering.GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("Metering.GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("Metering.GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),
        Sensor("Metering.GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("Metering.GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("Metering.GridMs.TotPF", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWhIn.Bat", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWhIn.ChaSta", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWhIn", Identifier.metering_total_absorbed, factor=1000, unit="kWh"),
        # Sensor("Metering.GridMs.TotWhIn", Identifier.metering_total_yield, factor=1000, unit="kWh"),
        Sensor("Metering.GridMs.TotWhOut.Bat", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWhOut", Identifier.metering_total_consumption, factor=1000, unit="kWh"),
        Sensor("Metering.GridMs.TotWIn.ChaSta", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWIn", Identifier.metering_power_absorbed, factor=1, unit="W"),
        Sensor("Metering.GridMs.TotWOut", Identifier.metering_power_supplied, factor=1, unit="W"),
        Sensor("Metering.GridMs.VA.phsA", Identifier.grid_apparent_power_l1, factor=1, unit="VA"),
        Sensor("Metering.GridMs.VA.phsB", Identifier.grid_apparent_power_l2, factor=1, unit="VA"),
        Sensor("Metering.GridMs.VA.phsC", Identifier.grid_apparent_power_l3, factor=1, unit="VA"),
        Sensor("Metering.GridMs.VAr.phsA", Identifier.grid_reactive_power_l1, factor=1, unit="var"),
        Sensor("Metering.GridMs.VAr.phsB", Identifier.grid_reactive_power_l2, factor=1, unit="var"),
        Sensor("Metering.GridMs.VAr.phsC", Identifier.grid_reactive_power_l3, factor=1, unit="var"),
        Sensor("Metering.GridMs.W.phsA", Identifier.power_l1, factor=1, unit="W"),
        Sensor("Metering.GridMs.W.phsB", Identifier.power_l2, factor=1, unit="W"),
        Sensor("Metering.GridMs.W.phsC", Identifier.power_l3, factor=1, unit="W"),
        Sensor("Metering.TotFeedTms", None),
        Sensor("Metering.TotFeedTms", None, factor=1, unit=None),
        Sensor("Metering.TotOpTms", None),
        Sensor("Metering.TotOpTms", None, factor=1, unit=None),
        Sensor("Metering.TotWhIn", None, factor=1, unit=None),
        Sensor("Metering.TotWhOut", Identifier.total_yield, factor=1000, unit="kWh"),
        Sensor("Metering.TotWhOut.Pv", Identifier.pv_gen_meter, factor=1000, unit="kWh"),
        Sensor("Nameplate.CmpMain.HwRev", None, factor=1, unit=None),
        Sensor("Nameplate.CmpMain.Rev", None, factor=1, unit=None),
        Sensor("Nameplate.CmpOS.SwRev", None, factor=1, unit=None),
        Sensor("Nameplate.Location", None, factor=1, unit=None),
        Sensor("Nameplate.MainModel", None, factor=1, unit=None),
        Sensor("Nameplate.Model", None, factor=1, unit=None),
        Sensor("Nameplate.ModelStr", None, factor=1, unit=None),
        Sensor("Nameplate.PkgRev", None, factor=1, unit=None),
        Sensor("Nameplate.SerNum", None, factor=1, unit=None),
        Sensor("Nameplate.Vendor", None, factor=1, unit=None),
        Sensor("Operation.Bat.Health", None, factor=1, unit=None),
        Sensor("Operation.BckStt", None, factor=1, unit=None),
        Sensor("Operation.DrtStt", None, factor=1, unit=None),
        Sensor("Operation.EVeh.ChaStt", Identifier.operating_status, factor=1, unit=None, mapper=SMATagList),
        Sensor("Operation.EVeh.Health", Identifier.status, factor=1, unit=None, mapper=SMATagList),
        Sensor("Operation.EvtCntIstl", None, factor=1, unit=None),
        Sensor("Operation.EvtCntUsr", None, factor=1, unit=None),
        Sensor("Operation.Evt.Dsc", None, factor=1, unit=None),
        Sensor("Operation.Evt.EvtNo", None, factor=1, unit=None),
        Sensor("Operation.Evt.Msg", None, factor=1, unit=None),
        Sensor("Operation.GriSwCnt", None, factor=1, unit=None),
        Sensor("Operation.GriSwStt", Identifier.grid_relay_status, unit=None, mapper=SMATagList),
        Sensor("Operation.Health", Identifier.status, factor=1, unit=None, mapper=SMATagList),
        Sensor("Operation.Health", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Alm", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Ok", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Wrn", None, factor=1, unit=None),
        Sensor("Operation.OpStt", Identifier.operating_status_genereal, unit=None, mapper=SMATagList),
        Sensor("Operation.PvGriConn", None, factor=1, unit=None),
        Sensor("Operation.RstrLokStt", None, factor=1, unit=None),
        Sensor("Operation.RunStt", None, factor=1, unit=None),
        Sensor("Operation.StandbyStt", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModAct", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModStt", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimNom", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimSrc", None, factor=1, unit=None),
        Sensor("Operation.WMinLimSrc", None, factor=1, unit=None),
        Sensor("Portal.EnnexOS.ConnStt", None, factor=1, unit=None),
        Sensor("PvGen.PvWh", None, factor=1, unit=None),
        Sensor("PvGen.PvW", None, factor=1, unit=None),
        Sensor("Setpoint.PlantControl.CmpBMS.OpMod.1", None, factor=1, unit=None),
        Sensor("Setpoint.PlantControl.InOut.GO1", None, factor=1, unit=None),  # codespell:ignore
        Sensor("Spdwr.ActlDnsSrvIp", None, factor=1, unit=None),
        Sensor("Spdwr.ActlGwIp", None, factor=1, unit=None),
        Sensor("Spdwr.ActlIp", None, factor=1, unit=None),
        Sensor("Spdwr.ActlSnetMsk", None, factor=1, unit=None),
        Sensor("Spdwr.AutoCfgIsOn", None, factor=1, unit=None),
        Sensor("Spdwr.ComSocA.Stt", None, factor=1, unit=None),
        Sensor("Spdwr.IgmpQryTms", None, factor=1, unit=None),
        Sensor("Spdwr.IgmpQryTx", None, factor=1, unit=None),
        Sensor("SunSpecSig.SunSpecTx.1", None, factor=1, unit=None),
        Sensor("SwCmp.CmpEnnexOS.Frwk.SwRev", None, factor=1, unit=None),
        Sensor("Sys.DevRstr", None, factor=1, unit=None),
        Sensor("Upd.AutoUpdIsOn", None, factor=1, unit=None),
        Sensor("Upd.Stt", None, factor=1, unit=None),
        Sensor("WebConn.IsOn", None, factor=1, unit=None),
        Sensor("WebConn.Stt", None, factor=1, unit=None),
        Sensor("WebConn.SwRev", None, factor=1, unit=None),
        Sensor("Wl.AcqStt", None, factor=1, unit=None),
        Sensor("Wl.AntMod", None, factor=1, unit=None),
        Sensor("Wl.ConnStt", None, factor=1, unit=None),
        Sensor("Wl.SigPwr", None, factor=1, unit=None),
        Sensor("Wl.SoftAcsConnStt", None, factor=1, unit=None),

        Sensor("Operation.Dmd.WCtl", Identifier.active_power_limitation, factor=1, unit="W"),
        Sensor("InOut.DigIn.CurWCtlNom", Identifier.power_limit_via_io, factor=1, unit="%"),  # codespell:ignore
        Sensor("Inverter.WMaxIn", Identifier.power_limit_in, factor=1, unit="W")
]
# fmt: on
name2sensor = {i.key: i for i in enneoxSensors}


def getSensorForDevice(productTagId: int) -> tuple[list[Sensor], list[str]] | None:
    """Return the profile for a device"""
    for profil in ennexosSensorProfiles:
        if productTagId in profil[0]:
            unknown: list[str] = []
            expected_sensors: list[Sensor] = []
            for sName in profil[1]:
                sensor = name2sensor.get(sName, None)
                if sensor:
                    expected_sensors.append(sensor)
                else:
                    unknown.append(sName)
            return (expected_sensors, unknown)
    return None



"""Implementation for SMA Speedwire

Originally based on https://github.com/Wired-Square/sma-query/blob/main/src/sma_query_sw/commands.py
Improved with Information from https://github.com/mhop/fhem-mirror/blob/master/fhem/FHEM/76_SMAInverter.pm
Receiver classes completely reimplemented by little.yoda
"""

# pylint: disable=too-many-lines

import ctypes
import time
from ctypes import LittleEndianStructure
from typing import Annotated, Any, Dict

import dataclasses_struct as dcs

from .const import Identifier, SMATagList
from .sensor import Sensor

responseDef: dict[str, list[dict[str, Any]]] = {
    "00464B01": [],  # Netzspannung Phase L1 gegen L2
    "00464C01": [],  # Netzspannung Phase L2 gegen L3
    "00464D01": [],  # Netzspannung Phase L3 gegen L1
    "00416601": [],  # WaitingTimeUntilFeedIn  WaitingTimeUntilFeedIn
    "00411E01": [],  # MaxACPower
    "00411F01": [],  # MaxACPower
    "00412001": [],  # MaxACPower
    "00462E01": [],  # OperationTime
    "00462F01": [],  # OperationTime
    "00823401": [
        {
            "cmd": "Firmware",
            "format": "version",
            "sensor": Sensor("Firmware", Identifier.device_sw_version),
            "idx": 4,
            "overwrite": False,
        }
    ],
    "00260101": [
        {
            "cmd": "EnergyProduction",
            "format": "uint",
            "sensor": Sensor("total", Identifier.total_yield, factor=1000, unit="kWh"),
            "idx": 0,
        }
    ],
    "40251E01": [
        {
            "cmd": "SpotDCPower",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_power1", Identifier.pv_power_a, factor=1, unit="W"
            ),
            "idx": 0,
        }
    ],
    "40251E02": [
        {
            "cmd": "SpotDCPower",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_power2", Identifier.pv_power_b, factor=1, unit="W"
            ),
            "idx": 0,
        }
    ],
    "40251E03": [
        {
            "cmd": "SpotDCPower",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_power3", Identifier.pv_power_c, factor=1, unit="W"
            ),
            "idx": 0,
        }
    ],
    "40464001": [
        {
            "cmd": "spot_ac_power",
            "format": "int",
            "sensor": Sensor("spot_ac_power1", Identifier.power_l1, factor=1, unit="W"),
            "idx": 0,
        }
    ],
    "40464101": [
        {
            "cmd": "spot_ac_power",
            "format": "int",
            "sensor": Sensor("spot_ac_power2", Identifier.power_l2, factor=1, unit="W"),
            "idx": 0,
        }
    ],
    "40464201": [
        {
            "cmd": "spot_ac_power",
            "format": "int",
            "sensor": Sensor("spot_ac_power3", Identifier.power_l3, factor=1, unit="W"),
            "idx": 0,
        }
    ],
    "08412801": [
        {
            "cmd": "OperatingStatus",
            "format": "uint",
            "sensor": [
                Sensor(
                    "GeneralOperatingStatus",
                    Identifier.operating_status_genereal,
                    factor=1,
                    mapper=SMATagList,
                ),
                Sensor(
                    "GeneralOperatingStatus2",
                    "Temp GeneralOperatingStatus2",
                    factor=1,
                    mapper=SMATagList,
                ),
            ],
            "idx": 0xFF,
        }
    ],
    "0046C201": [
        {
            "cmd": "SpotDCPower_3",
            "format": "uint",
            "sensor": Sensor("pv_power", Identifier.pv_power, unit="W"),
            "idx": 0,
        }
    ],
    "40263F01": [
        {
            "cmd": "SpotACTotalPower",
            "format": "int",
            "sensor": Sensor("grid_power", Identifier.grid_power, unit="W"),
            "idx": 0,
        }
    ],
    "00464801": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "spot_ac_voltage1", Identifier.voltage_l1, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "00464901": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "ac_voltage_l1", Identifier.voltage_l2, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "00464A01": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "ac_voltage_l2", Identifier.voltage_l3, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "40465301": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "ac_current_l1", Identifier.current_l1, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "40465401": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "ac_current_l2", Identifier.current_l2, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "40465501": [
        {
            "cmd": "spot_ac_voltage",
            "format": "uint",
            "sensor": Sensor(
                "ac_current_l3", Identifier.current_l3, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "08412B01": [
        {
            "cmd": "OperatingStatus",
            "format": "uint",
            "sensor": Sensor(
                "OperatingStatus", Identifier.operating_status, mapper=SMATagList
            ),
            "idx": 0xFF,
        }
    ],
    "0046C301": [
        {
            "cmd": "PVEnergyProduction",
            "format": "uint",
            "sensor": Sensor(
                "pvEnergyProduction", Identifier.pv_gen_meter, unit="kWh", factor=1000
            ),
            "idx": 0,
        }
    ],
    "0846A601": [
        {
            "cmd": "GridConection",
            "format": "uint",
            #                "sensor": , TODO
            "idx": 0xFF,
        }
    ],
    "08214801": [
        {
            "cmd": "DeviceStatus",
            "format": "uint",
            "sensor": Sensor("inverter_status", Identifier.status, mapper=SMATagList),
            "idx": 0xFF,
        }
    ],
    "40451F01": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_voltage1", Identifier.pv_voltage_a, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "40451F02": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_voltage2", Identifier.pv_voltage_b, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "40451F03": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_voltage3", Identifier.pv_voltage_c, factor=100, unit="V"
            ),
            "idx": 0,
        }
    ],
    "40452101": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_current1", Identifier.pv_current_a, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "40452102": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_current2", Identifier.pv_current_b, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "40452103": [
        {
            "cmd": "SpotACCurrent",
            "format": "uint",
            "sensor": Sensor(
                "spot_dc_current3", Identifier.pv_current_c, factor=1000, unit="A"
            ),
            "idx": 0,
        }
    ],
    "08416401": [
        {
            "cmd": "GridRelayStatus",
            "format": "uint",
            "sensor": Sensor(
                "GridRelayStatus", Identifier.grid_relay_status, mapper=SMATagList
            ),
            "idx": 0xFF,
        }
    ],
    "00465701": [
        {
            "cmd": "SpotGridFrequency",
            "format": "uint",
            "sensor": Sensor(
                "grid_frequency", Identifier.frequency, factor=100, unit="Hz"
            ),
            "idx": 0,
        }
    ],
    "10821E01": [
        {
            "cmd": "TypeLabel",
            "format": "uint",
            # "sensor": TODO Vermutlich String
            "idx": 0,
        }
    ],
    "08821F01": [
        {
            "cmd": "TypeLabel",
            "format": "uint",
            "sensor": Sensor(
                "inverter_class", Identifier.device_class, mapper=SMATagList
            ),
            "idx": 0xFF,
            "overwrite": False,
        }
    ],
    "08822001": [
        {
            "cmd": "TypeLabel",
            "format": "uint",
            "sensor": Sensor(
                "inverter_type", Identifier.device_type, mapper=SMATagList
            ),
            "idx": 0xFF,
            "overwrite": False,
        }
    ],
    "00262201": [
        {
            "cmd": "energy_production",
            "format": "uint",
            "sensor": Sensor("today", Identifier.daily_yield, unit="Wh"),
            "idx": 0,
        }
    ],
    "00254F01": [
        {
            "cmd": "",
            "format": "uint",
            "sensor": Sensor(
                "Insulation_2",
                Identifier.pv_isolation_resistance,
                unit="kOhm",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "40254E01": [
        {
            "cmd": "",
            "format": "uint",
            "sensor": Sensor(
                "Insulation_1",
                Identifier.insulation_residual_current,
                unit="mA",
            ),
            "idx": 0,
        }
    ],
    "00462401": [
        {
            "cmd": "EM_1",
            "format": "uint",
            "sensor": Sensor(
                "meter_yield",
                Identifier.metering_total_yield,
                unit="kWh",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "40237701": [
        {
            "cmd": "InverterTemperature",
            "format": "uint",
            "sensor": Sensor(
                "InverterTemperature",
                Identifier.temp_a,
                unit="°C",
                factor=100,
            ),
            "idx": 0,
        }
    ],
    "00295A01": [
        {
            "cmd": "ChargeStatus",
            "format": "uint",
            "sensor": Sensor("ChargeStatus", Identifier.battery_soc_total, unit="%"),
            "idx": 0,
        }
    ],
    "40491E01": [
        {
            "cmd": "BatteryInfo",
            "format": "uint",
            "sensor": Sensor("TEMP BAT_CYCLES", "TEMP Battery Cycles"),
            "idx": 0,
        }
    ],
    "40495B01": [
        {
            "cmd": "batteryTemp",
            "format": "uint",
            "sensor": Sensor(
                "batteryTemp",
                Identifier.battery_temp_a,
                unit="°C",
                factor=10,
            ),
            "idx": 0,
        }
    ],
    "00496701": [
        {
            "cmd": "SpotBatteryLoad",
            "format": "uint",
            "sensor": Sensor(
                "SpotBatteryLoad",
                Identifier.battery_charge_total,
                unit="kWh",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "00496801": [
        {
            "cmd": "SpotBatteryUnload",
            "format": "uint",
            "sensor": Sensor(
                "SpotBatteryUnLoad",
                Identifier.battery_discharge_total,
                unit="kWh",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "00495C01": [
        {
            "cmd": "battery_voltage_a",
            "format": "uint",
            "sensor": Sensor(
                "battery_voltage_a",
                Identifier.battery_voltage_a,
                unit="V",
                factor=100,
            ),
            "idx": 0,
        }
    ],
    "40495D01": [
        {
            "cmd": "battery_current_a",
            "format": "int",
            "sensor": Sensor(
                "battery_current_a",
                Identifier.battery_current_a,
                unit="A",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "00696E01": [
        {
            "cmd": "BatteryInfo_Capacity",
            "format": "uint",
            "sensor": Sensor(
                "BatteryInfo_Capacity", Identifier.battery_capacity_total, unit="%"
            ),
            "idx": 0,
        }
    ],
    "00496901": [
        {
            "cmd": "BatteryInfo_Charge",
            "format": "uint",
            "sensor": Sensor(
                "BatteryInfo_Charge", Identifier.battery_power_charge_total, unit="W"
            ),
            "idx": 0,
        }
    ],
    "00496A01": [
        {
            "cmd": "BatteryInfo_Charge",
            "format": "uint",
            "sensor": Sensor(
                "BatteryInfo_DisCharge",
                Identifier.battery_power_discharge_total,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "00462501": [
        {
            "cmd": "EM_1",
            "format": "uint",
            "sensor": Sensor(
                "EM1_1 Identifier.metering_total_absorbed",
                Identifier.metering_total_absorbed,
                unit="kWh",
                factor=1000,
            ),
            "idx": 0,
        }
    ],
    "40463601": [
        {
            "cmd": "EM_3",
            "format": "uint",
            "sensor": Sensor(
                "EM3 Identifier.metering_power_supplied",
                Identifier.metering_power_supplied,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "40463701": [
        {
            "cmd": "EM_3",
            "format": "uint",
            "sensor": Sensor(
                "EM3 Identifier.metering_power_absorbed",
                Identifier.metering_power_absorbed,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "00893701": [
        {
            "cmd": "BatteryInfo_4",
            "format": "uint",
            "sensor": Sensor(
                "EM3 BatteryInfo_4", "TEMP battery_capacity_rated", unit="Wh"
            ),
            "idx": 4,
        }
    ],
    "08412509": [
        {
            "cmd": "BackupRelayStatus",
            "format": "uint",
            "sensor": Sensor(
                "BackupRelayStatus", "TEMP_BackupRelayStatus", mapper=SMATagList
            ),
            "idx": 0xFF,
        }
    ],
    # "00469109": [{ # Meter_Grid_FeedIn
    #             "cmd": "EM_2",
    #             "format": "uint",
    #             "sensor": Sensor("EM_2", Identifier., unit="kWh", factor=1000),
    #             "idx": 0
    # }],
    # "00469209": [{ # Meter_Grid_FeedIn
    #             "cmd": "EM_2",
    #             "format": "uint",
    #             "sensor": Sensor("EM_2", Identifier.battery_discharge_total, unit="kWh", factor=1000),
    #             "idx": 0
    # }],
    "0046E809": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_feed_l1",
                Identifier.metering_active_power_feed_l1,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "0046E909": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_feed_l2",
                Identifier.metering_active_power_feed_l2,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "0046EA09": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_feed_l3",
                Identifier.metering_active_power_feed_l3,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "0046EB09": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_draw_l1",
                Identifier.metering_active_power_draw_l1,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "0046EC09": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_draw_l2",
                Identifier.metering_active_power_draw_l2,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "0046ED09": [
        {
            "cmd": "EM_4",
            "format": "uint",
            "sensor": Sensor(
                "active_power_draw_l3",
                Identifier.metering_active_power_draw_l3,
                unit="W",
            ),
            "idx": 0,
        }
    ],
    "40574609": [
        {
            "cmd": "SpotACCurrentBackup",
            "format": "uint",
            # "sensor": ,
            "idx": 0,
        }
    ],
    "40574709": [
        {
            "cmd": "SpotACCurrentBackup",
            "format": "uint",
            # "sensor": ,
            "idx": 0,
        }
    ],
    "40574809": [
        {
            "cmd": "SpotACCurrentBackup",
            "format": "uint",
            # "sensor": ,
            "idx": 0,
        }
    ],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
    # "": [{
    #             "cmd": "",
    #             "format": "uint",
    #             "sensor": ,
    #             "idx": 0
    # }],
}

commands: Dict[str, Dict[str, Any]] = {
    "login": {
        "command": 0xFFFD040C,
        "response": 0xFFFD040D,
        "first": 0x0030CB00,
        "last": 0x0030CB00,
    },
    "logoff": {"command": 0xFFFD010E, "response": 0xFFFD010F},
    "TypeLabel": {
        "command": 0x58000200,
        "first": 0x00821E00,
        "last": 0x008220FF,
    },
    "EnergyProduction": {
        "command": 0x54000200,
        "first": 0x00260100,
        "last": 0x002622FF,
    },
    "PVEnergyProduction": {
        "command": 0x54000200,
        "first": 0x0046C300,
        "last": 0x0046C3FF,
    },
    "SpotDCPower": {
        "command": 0x53800200,
        "first": 0x00251E00,
        "last": 0x00251EFF,
    },
    "SpotDCPower_3": {
        "command": 0x51000200,
        "first": 0x0046C200,
        "last": 0x0046C2FF,
    },
    "SpotACTotalPower": {
        "command": 0x51000200,
        "first": 0x00263F00,
        "last": 0x00263FFF,
    },
    "ChargeStatus": {
        "command": 0x51000200,
        "first": 0x00295A00,
        "last": 0x00295AFF,
    },
    "SpotDCVoltage": {
        "command": 0x53800200,
        "first": 0x00451F00,
        "last": 0x004521FF,
    },
    "SpotACCurrentBackup": {
        "command": 0x51000200,
        "first": 0x40574600,
        "last": 0x405748FF,
    },
    "BatteryInfo_TEMP": {
        "command": 0x51000200,
        "first": 0x00495B00,
        "last": 0x00495B10,
    },
    "BatteryInfo_UDC": {
        "command": 0x51000200,
        "first": 0x00495C00,
        "last": 0x00495C10,
    },
    "BatteryInfo_IDC": {
        "command": 0x51000200,
        "first": 0x00495D00,
        "last": 0x00495D10,
    },
    "BatteryInfo_Charge": {
        "command": 0x51000200,
        "first": 0x00496900,
        "last": 0x00496AFF,
    },
    "BatteryInfo_Capacity": {
        "command": 0x51000200,
        "first": 0x00696E00,
        "last": 0x00696E10,
    },
    "BatteryInfo": {
        "command": 0x51000200,
        "first": 0x00491E00,
        "last": 0x00495DFF,
    },
    "BatteryInfo_3": {
        "command": 0x58020200,
        "first": 0x08822C00,
        "last": 0x08822CFF,
    },
    "BatteryInfo_4": {
        "command": 0x58000200,
        "first": 0x00893700,
        "last": 0x008937FF,
    },
    "BatteryInfo_5": {
        "command": 0x58020200,
        "first": 0x00B18900,
        "last": 0x00B189FF,
    },
    "SpotGridFrequency": {
        "command": 0x51000200,
        "first": 0x00465700,
        "last": 0x004657FF,
    },
    "OperationTime": {
        "command": 0x54000200,
        "first": 0x00462E00,
        "last": 0x00462FFF,
    },
    "InverterTemperature": {
        "command": 0x52000200,
        "first": 0x00237700,
        "last": 0x002377FF,
        "format": "int",
    },
    "MaxACPower": {
        "command": 0x51000200,
        "first": 0x00411E00,
        "last": 0x004120FF,
    },
    "MaxACPower2": {
        "command": 0x51000200,
        "first": 0x00832A00,
        "last": 0x00832AFF,
    },
    "GridRelayStatus": {
        "command": 0x51800200,
        "first": 0x00416400,
        "last": 0x004164FF,
    },
    "BackupRelayStatus": {
        "command": 0x51800200,
        "first": 0x08412500,
        "last": 0x084125FF,
    },
    "GridConection": {
        "command": 0x51800200,
        "first": 0x0046A600,
        "last": 0x0046A6FF,
    },
    "OperatingStatus": {
        "command": 0x51800200,
        "first": 0x08412B00,
        "last": 0x08412BFF,
    },
    "GeneralOperatingStatus": {
        "command": 0x51800200,
        "first": 0x00412800,
        "last": 0x004128FF,
    },
    "WaitingTimeUntilFeedIn": {
        "command": 0x51000200,
        "first": 0x00416600,
        "last": 0x004166FF,
    },
    "DeviceStatus": {
        "command": 0x51800200,
        "first": 0x00214800,
        "last": 0x002148FF,
    },
    "SpotBatteryLoad": {
        "command": 0x54000200,
        "first": 0x00496700,
        "last": 0x004967FF,
    },
    "SpotBatteryUnload": {
        "command": 0x54000200,
        "first": 0x00496800,
        "last": 0x004968FF,
    },
    "EM_1": {
        "command": 0x54000200,
        "first": 0x00462400,
        "last": 0x004628FF,
    },
    "EM_2": {
        "command": 0x54000200,
        "first": 0x40469100,
        "last": 0x404692FF,
    },
    "EM_3": {
        "command": 0x51000200,
        "first": 0x00463600,
        "last": 0x004637FF,
    },
    "EM_4": {
        "command": 0x51000200,
        "first": 0x0046E800,
        "last": 0x0046EDFF,
    },
    "Insulation_1": {
        "command": 0x51020200,
        "first": 0x00254E00,
        "last": 0x00254FFF,
    },
    "Insulation_2": {
        "command": 0x51020200,
        "first": 0x00254F00,
        "last": 0x00254FFF,
    },
    "Firmware": {
        "command": 0x58000200,
        "first": 0x00823400,
        "last": 0x008234FF,
    },
    "energy_production": {
        "command": 0x54000200,
        "first": 0x00260100,
        "last": 0x002622FF,
    },
    "spot_ac_power": {
        "command": 0x51000200,
        "first": 0x00464000,
        "last": 0x004642FF,
    },
    "spot_ac_voltage": {
        "command": 0x51000200,
        "first": 0x00464800,
        "last": 0x004655FF,
    },
}


@dcs.dataclass(dcs.BIG_ENDIAN)
class speedwireHeader:
    """Speedwire header"""

    sma: Annotated[bytes, 4]
    tag42_length: dcs.U16
    tag42_tag0x02A0: dcs.U16
    group1: dcs.U32
    smanet2_length: dcs.U16
    smanet2_tagID: dcs.U16
    protokoll: dcs.U16

    def check6065(self) -> bool:
        """Check for 6065 Type, used by inverters. Size is not checked at this stage."""
        return (
            self.sma == b"SMA\x00"
            and self.tag42_length == 4
            and self.tag42_tag0x02A0 == 0x02A0
            and self.group1 == 1
            and self.smanet2_tagID == 0x10
            and self.protokoll == 0x6065
        )

    def check6069(self) -> bool:
        """Check for 6069 Type, used by energymeters.  Size is not checked at this stage."""
        return (
            self.sma == b"SMA\x00"
            and self.tag42_length == 4
            and self.tag42_tag0x02A0 == 0x02A0
            and self.group1 == 1
            and self.smanet2_tagID == 0x10
            and self.protokoll == 0x6069
        )

    def __str__(self) -> str:
        """customized output. Use hex-format for important values."""
        return f"speedwireHeader(sma:{self.sma!r} tag42_length:{self.tag42_length} tag42_tag0x02A0:{self.tag42_tag0x02A0:#04x} group1:{self.group1} smanet2_length:{self.smanet2_length} smanet2_tagID:{self.smanet2_tagID:#02x} protokoll:{self.protokoll:#04x})"

    def isDiscoveryResponse(self) -> bool:
        """Check if this message is a response to a discovery request."""
        return (
            self.sma == b"SMA\x00"
            and self.tag42_length == 4
            and self.tag42_tag0x02A0 == 0x02A0
            and self.group1 == 1
            and self.smanet2_length == 2
            and self.smanet2_tagID == 0
            and self.protokoll == 1
        )


@dcs.dataclass(dcs.BIG_ENDIAN)
class speedwireData2Tag:
    smanet2_lengthPayload: dcs.U16
    smanet2_id: dcs.U16
    protokoll: dcs.U16


@dcs.dataclass(dcs.LITTLE_ENDIAN)
class speedwireHeader6065:
    """Speedwire Header2 for 6065 Messages."""

    # https://github.com/RalfOGit/libspeedwire

    #        13 Bytes     0x26/38       00106065 09A0
    # $cmd = $cmdheader . $pktlength . $esignature . $target_ID . "0000" . $myID . "0000" . "00000000" . $spkt_ID . $cmd_ID . "00000000";

    unknown09A0E0: Annotated[bytes, 2]
    dest_susyid: dcs.U16
    dest_serial: dcs.U32
    dest_control: dcs.U16
    src_susyid: dcs.U16
    src_serial: dcs.U32
    src_control: dcs.U16
    error: dcs.U16
    fragment: dcs.U16

    pktId: dcs.U16
    cmdid: dcs.U32
    firstRegister: dcs.U32
    lastRegister: dcs.U32

    def isLoginResponse(self) -> bool:
        """Check if this message is a response to a login request."""
        return self.cmdid == 0xFFFD040D

    def __str__(self) -> str:
        """customized output. Use hex-format for important values."""
        return f"speedwireHeader6065(?:{self.unknown09A0E0.hex()} Src (ID,SNR,CNT): {self.src_susyid} {self.src_serial} {self.src_control} Dest (ID,SNR,CNT): {self.dest_susyid} {self.dest_serial} {self.dest_control}   error:{self.error} fragment:{self.fragment} pktId:{self.pktId} cmdid:{self.cmdid:#010x} firstRegister:{self.firstRegister:#010x} lastRegister:{self.lastRegister:#010x})"


@dcs.dataclass(dcs.BIG_ENDIAN)
class speedwireHeader6069:
    """Speedwire Header2 for 6069 Messages. 10 Bytes"""

    src_susyid: dcs.U16
    src_serial: dcs.U32

    timestamp: dcs.U32


@dcs.dataclass(dcs.LITTLE_ENDIAN)
class speedwireHeader6065x010:
    pass


# Originally based on https://github.com/Wired-Square/sma-query/blob/main/src/sma_query_sw/commands.py
class SpeedwireFrame:
    """Class for the send speedwire messages"""

    APP_ID = 125
    ANY_SERIAL = 0xFFFFFFFF
    ANY_SUSYID = 0xFFFF

    # Login Timeout in seconds
    LOGIN_TIMEOUT = 900

    def get_encoded_pw(self, password: str, installer: bool = False) -> bytearray:
        """Encodes the password"""
        byte_password = bytearray(password.encode("ascii"))

        if installer:
            login_code = 0xBB
        else:
            login_code = 0x88

        encodedpw = bytearray(12)

        for index in range(0, 12):
            if index < len(byte_password):
                encodedpw[index] = (login_code + byte_password[index]) % 256
            else:
                encodedpw[index] = login_code

        return encodedpw

    _frame_sequence = 1
    _id = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"SMA\x00"))
    _tag0 = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x04\x02\xA0"))
    _group1 = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x00\x00\x01"))
    _eth_sig = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x10\x60\x65"))
    _ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))
    _ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))

    _data_length = 0  # Placeholder value
    _longwords = 0  # Placeholder value
    _ctrl = 0  # Placeholder value

    class FrameHeader(LittleEndianStructure):
        """Frame Header"""

        _pack_ = 1
        _fields_ = [
            ("id", ctypes.c_ubyte * 4),
            ("tag0", ctypes.c_ubyte * 4),
            ("group1", ctypes.c_ubyte * 4),
            ("data_length", ctypes.c_uint16),
            ("eth_sig", ctypes.c_ubyte * 4),
            ("longwords", ctypes.c_ubyte),
            ("ctrl", ctypes.c_ubyte),
        ]

    class DataHeader(LittleEndianStructure):
        # pylint: disable=too-few-public-methods
        """Data header"""
        _pack_ = 1
        _fields_ = [
            ("dst_sysyid", ctypes.c_uint16),
            ("dst_serial", ctypes.c_uint32),
            ("ctrl2_1", ctypes.c_ubyte * 2),
            ("app_id", ctypes.c_uint16),
            ("app_serial", ctypes.c_uint32),
            ("ctrl2_2", ctypes.c_ubyte * 2),
            ("preamble", ctypes.c_uint32),
            ("sequence", ctypes.c_uint16),
        ]

    class LogoutFrame(LittleEndianStructure):
        # pylint: disable=too-few-public-methods
        """Logout"""
        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("data_start", ctypes.c_uint32),
            ("data_end", ctypes.c_uint32),
        ]

    class LoginFrame(LittleEndianStructure):
        # pylint: disable=too-few-public-methods
        """Login"""
        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("login_type", ctypes.c_uint32),
            ("timeout", ctypes.c_uint32),
            ("time", ctypes.c_uint32),
            ("data_start", ctypes.c_uint32),
            ("user_password", ctypes.c_ubyte * 12),
            ("data_end", ctypes.c_uint32),
        ]

    class QueryFrame(LittleEndianStructure):
        # pylint: disable=too-few-public-methods
        """Query Frame"""

        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("first", ctypes.c_uint32),
            ("last", ctypes.c_uint32),
            ("data_end", ctypes.c_uint32),
        ]

    # def getLogoutFrame(self, inverter):
    #     frame_header = self.getFrameHeader()
    #     frame_data_header = self.getDataHeader(inverter)
    #     frame_data = self.LogoutFrame()

    #     frame_header.ctrl = 0xA0
    #     frame_data_header.dst_sysyid = 0xFFFF
    #     frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x03"))
    #     frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x03"))

    #     frame_data.command = commands["logoff"]["command"]
    #     frame_data.data_start = 0xFFFFFFFF
    #     frame_data.data_end = 0x00000000

    #     data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

    #     frame_header.data_length = int.from_bytes(data_length.to_bytes(2, "big"), "little")

    #     frame_header.longwords = (data_length // 4)

    #     return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getLoginFrame(self, password: str, serial: int, installer: bool) -> bytes:
        # pylint: disable=too-few-public-methods
        """Returns a Login Frame"""
        frame_header = self.getFrameHeader()
        frame_data_header = self.getDataHeader(password, serial)
        frame_data = self.LoginFrame()

        frame_header.ctrl = 0xA0
        frame_data_header.dst_sysyid = 0xFFFF
        frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(
            bytearray(b"\x00\x01")
        )
        frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(
            bytearray(b"\x00\x01")
        )

        frame_data.command = commands["login"]["command"]
        frame_data.login_type = (0x07, 0x0A)[installer]
        frame_data.timeout = self.LOGIN_TIMEOUT
        frame_data.time = int(time.time())
        frame_data.data_start = 0x00000000  # Data Start
        frame_data.user_password = (ctypes.c_ubyte * 12).from_buffer(
            self.get_encoded_pw(password, installer)
        )
        frame_data.date_end = 0x00000000  # Packet End

        data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

        frame_header.data_length = int.from_bytes(
            data_length.to_bytes(2, "big"), "little"
        )

        frame_header.longwords = data_length // 4

        return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getQueryFrame(self, password: str, serial: int, command_name: str) -> bytes:
        """Return Query Frame"""
        frame_header = self.getFrameHeader()
        frame_data_header = self.getDataHeader(password, serial)
        frame_data = self.QueryFrame()

        command = commands[command_name]

        frame_header.ctrl = 0xA0
        frame_data_header.dst_sysyid = 0xFFFF
        frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(
            bytearray(b"\x00\x00")
        )
        frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(
            bytearray(b"\x00\x00")
        )

        frame_data.command = command["command"]
        frame_data.first = command["first"]
        frame_data.last = command["last"]
        frame_data.date_end = 0x00000000

        data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

        frame_header.data_length = int.from_bytes(
            data_length.to_bytes(2, "big"), "little"
        )

        frame_header.longwords = data_length // 4

        return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getFrameHeader(self) -> FrameHeader:
        """Return Frame Header"""
        newFrameHeader = self.FrameHeader()
        newFrameHeader.id = self._id
        newFrameHeader.tag0 = self._tag0
        newFrameHeader.group1 = self._group1
        newFrameHeader.data_length = self._data_length
        newFrameHeader.eth_sig = self._eth_sig
        newFrameHeader.longwords = self._longwords
        newFrameHeader.ctrl = self._ctrl

        return newFrameHeader

    def getDataHeader(self, password: str, serial: int) -> DataHeader:
        """Return Data Header"""
        newDataHeader = self.DataHeader()

        newDataHeader.dst_susyid = self.ANY_SUSYID
        newDataHeader.dst_serial = self.ANY_SERIAL
        newDataHeader.ctrl2_1 = self._ctrl2_1
        newDataHeader.app_id = self.APP_ID
        newDataHeader.app_serial = serial
        newDataHeader.ctrl2_2 = self._ctrl2_2
        newDataHeader.preamble = 0
        newDataHeader.sequence = self._frame_sequence | 0x8000

        self._frame_sequence += 1

        return newDataHeader



# flake8: noqa
sempxsd = """<?xml version="1.0" encoding="utf-8"?>
<!--
# SEMP - Simple Energy Management Protocol
#
# Version: 1.3.0 , 2015-03-24
#
# SMA Solar Technology AG
# 34266 Niestetal, Germany
-->
<xs:schema targetNamespace="http://www.sma.de/communication/schema/SEMP/v1"
    elementFormDefault="qualified"
    xmlns="http://www.sma.de/communication/schema/SEMP/v1"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
>
  <xs:element name="Device2EM">
    <xs:annotation>
      <xs:documentation>
        This data structure encapsulates all information transmitted by the device to the Energy Management System
      </xs:documentation>
    </xs:annotation>
    <xs:complexType>
      <xs:sequence>
        <xs:element name="DeviceInfo" type="DeviceInfoType" minOccurs="0" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>
              A DeviceInfo encapsulates all static information about a device such as identification, capabilities, restrictions, etc.
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:element name="DeviceStatus" type="DeviceStatusType"  minOccurs="0" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>
              A DeviceStatus encapsulates the status information of a device, i.e. all measurements and properties representing the current status of the device
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:element name="PlanningRequest" type="PlanningRequestType" minOccurs="0" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>
              A PlanningRequest allows specification of the needs of the device with regard to energy, running time or a certain behaviour of sensor values.
              As long as an energy need is pending, the gateway has to announce them in terms of PlanningRequests.
              PlanningRequests are not incremental, i.e. requests that sent to the EM in previous messages are discarded
              and replaced by the newest PlanningRequests. Missing timeframes or an empty PlanningRequest list are interpreted
              as if the energy needs are already satisfied.
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:element name="Messages" type="MessageListType" minOccurs="0" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>
              This element provides information about an error or event the EM should be informed about.
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:any namespace="##other" minOccurs="0" maxOccurs="unbounded" processContents="lax">
          <xs:annotation>
            <xs:documentation>
              Reserved for future use. Devices should ignore unknown elements.
            </xs:documentation>
          </xs:annotation>
        </xs:any>
      </xs:sequence>
    </xs:complexType>
  </xs:element>
  
  <xs:element name="EM2Device">
    <xs:annotation>
      <xs:documentation>
        This data structure encapsulates all information transmitted by the Energy Management System to the device
      </xs:documentation>
    </xs:annotation>
    <xs:complexType>
      <xs:sequence>
        <xs:element name="DeviceControl" type="DeviceControlType" minOccurs="0" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>
              This element contains load control recommendations for a device.
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:element name="Messages" type="MessageListType" minOccurs="0" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>
              This element provides information about an error or event the device should be informed about.
            </xs:documentation>
          </xs:annotation>
        </xs:element>
        <xs:any namespace="##other" minOccurs="0" maxOccurs="unbounded" processContents="lax">
          <xs:annotation>
            <xs:documentation>
              Reserved for future use. Devices should ignore unknown elements.
            </xs:documentation>
          </xs:annotation>
        </xs:any>
      </xs:sequence>
    </xs:complexType>
  </xs:element>

  <xs:simpleType name="DeviceIdType">
    <xs:annotation>
      <xs:documentation>
        Type for unique device IDs.
        A device ID consists of a vendor ID type (VID type), a vendor ID (VID), a serial number and sub-device ID.
        <p/>
        The vendor ID should be globally unique. For this purpose either IANA PEN or IEEE OUI (known from MAC addresses) should be used. Although it is possible to define non-globally unique vendor IDs this option must not be used for products.
        The kind of vendor ID is defined by the vendor ID type. The following types are defined:
        <ul>
          <li>0x0: IANA PEN (32 bit, Format: 0xXXXXXXXX)</li>
          <li>0x1: IEEE OUI (24 bit, Format: 0x00XXXXXX)</li>
          <li>0xF: non-unique ID in local address space (32 bit). IMPORTANT: Use only for demos, examples or testing, not for products.</li>
        </ul>
        <p/>
        The serial number can be freely defined by the vendor as it specific to the vendor's address space. It must be unique for each of the vendor's devices.
        <p/>
        The sub-device ID can be used to point out, that a physical device consists of multiple virtual devices. The virtual device with sub-device ID 0 should be the main device. This information is only used to group devices in GUIs at the moment.
        <p/>
        Format: [VID Type:4bit]-[VID:32bit]-[Serial:48bit]-[SubDev-ID:8bit]
        Example 1: "0-00008CAD-112233445566-00" (IANA PEN: 36013)
        Example 2: "F-11223344-112233445566-00" (local address, only for testing, demos, examples)
      </xs:documentation>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="[a-fA-F0-9]{1}-[a-fA-F0-9]{8}-[a-fA-F0-9]{12}-[a-fA-F0-9]{2}"/>
    </xs:restriction>
  </xs:simpleType>

  <xs:simpleType name="RelOrAbsTimeType">
    <xs:annotation>
      <xs:documentation>
        Type representing timestamps that can either be relative (seconds relative to the current point of time) or absolute (Unix timestamp UTC in seconds since 01.01.1970).
        The device specifies the interpretation globally with the DeviceInfo.Capabilities.Timestamps element.
        Devices that do not have a synchronized clock (with time server protocols like NTP or radio control like DCF77) or do not have a reliable absolute time source should use relative timestamps.
      </xs:documentation>
    </xs:annotation>
    <xs:restriction base="xs:long"/>
  </xs:simpleType>
  
  <xs:complexType name="MessageListType">
    <xs:annotation>
      <xs:documentation>
        This type encapsulates a list of messages.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Message" type="MessageType" minOccurs="1" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>
            This element encapsulates information about an error or event.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:simpleType name="MessageTypeRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="InvalidTimeframe"/>
      <xs:enumeration value="DeviceControlIgnored"/>
      <xs:enumeration value="Other"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="MessageLevelRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="Info"/>
      <xs:enumeration value="Warn"/>
      <xs:enumeration value="Error"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:complexType name="MessageType">
    <xs:annotation>
      <xs:documentation>
        This type encapsulates information about an error or event.
        It can also be used for data-tunnelling. In this case only the Type and Data elements might be set.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Type" type="xs:string" minOccurs="1" maxOccurs="1">
        <!-- ref=MessageTypeRefType -->
        <xs:annotation>
          <xs:documentation>
            The type of message used as a hint to interpret the message contents.
            See MessageTypeRefType for known values.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Level" type="xs:string" minOccurs="0" maxOccurs="1">
        <!-- ref=MessageLevelRefType -->
        <xs:annotation>
          <xs:documentation>
            The severity of the message.
            See MessageLevelRefType for known values ("Info", "Warn", "Error").
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Data" type="MessageDataType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Additional data describing the error or event (e.g. for automatic evaluation).
            For "InvalidTimeframe" the element DeviceId will be set.
            For "DeviceControlIgnored" the element DeviceId will be set.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Text" type="xs:string" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Description of the error/event.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="MessageDataType">
    <xs:annotation>
      <xs:documentation>
        Additional data describing the error/event.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeviceId" type="DeviceIdType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the device this message applies to.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Timestamp" type="RelOrAbsTimeType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Time of the occurrence of the error/event.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:any namespace="##other" minOccurs="0" maxOccurs="unbounded" processContents="lax">
        <xs:annotation>
          <xs:documentation>
            Reserved for future use. Devices should ignore unknown elements.
          </xs:documentation>
        </xs:annotation>
      </xs:any>
    </xs:sequence>
  </xs:complexType>

  <xs:complexType name="DeviceInfoType">
    <xs:annotation>
      <xs:documentation>
        A DeviceInfo encapsulates all static information about a device such as identification, capabilities, restrictions, etc.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Identification" type="IdentificationType">
        <xs:annotation>
          <xs:documentation>
            General information for identifying the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Characteristics" type="CharacteristicsType">
        <xs:annotation>
          <xs:documentation>
            Information on the characteristics of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Capabilities" type="CapabilitiesType">
        <xs:annotation>
          <xs:documentation>
            This element encapsulates information about the capabilities of the device
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:simpleType name="DeviceTypeRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="AirConditioning"/>
      <xs:enumeration value="Charger"/>
      <xs:enumeration value="DishWasher"/>
      <xs:enumeration value="Dryer"/>
      <xs:enumeration value="ElectricVehicle"/>
      <xs:enumeration value="EVCharger"/>
      <xs:enumeration value="Freezer"/>
      <xs:enumeration value="Fridge"/>
      <xs:enumeration value="Heater"/>
      <xs:enumeration value="HeatPump"/>
      <xs:enumeration value="Motor"/>
      <xs:enumeration value="Pump"/>
      <xs:enumeration value="WashingMachine"/>
      <xs:enumeration value="Other"/>
    </xs:restriction>
  </xs:simpleType> 
  <xs:complexType name="IdentificationType">
    <xs:annotation>
      <xs:documentation>
        General information for identifying the device.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeviceId" type="DeviceIdType">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="DeviceName" type="xs:string">
        <xs:annotation>
          <xs:documentation>
            Human readable device name
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="DeviceType" type="DeviceTypeRefType">
        <!-- ref=DeviceTypeRefType -->
        <xs:annotation>
          <xs:documentation>
            Human readable type of the device. See DeviceTypeRefType for well-known types.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="DeviceSerial" type="xs:string">
        <xs:annotation>
          <xs:documentation>
            Vendor specific serial number of the device
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="DeviceVendor" type="xs:string">
        <xs:annotation>
          <xs:documentation>
            Human readable name of the device vendor
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="DeviceURL" type="xs:string" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Configuration URL of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="CharacteristicsType">
    <xs:annotation>
      <xs:documentation>
        Information on the characteristics of the device.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="MaxPowerConsumption" type="xs:int">
        <xs:annotation>
          <xs:documentation>
            Maximum power consumption of the device in Watts. If the device is controllable with regard to power consumption, the recommendation
            of the energy management system will never exceed this value.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinPowerConsumption" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device power is controllable by the EM, this element determines the minimum power consumption of the device. Recommendations
            of the energy management system will never go below this value.

            If the device power is not controllable by the EM, do not provide this element - only provide the MaxPowerConsumption. 
            This applies to fixed power devices as well as devices that use varying power-levels but are controllable in power by the EM
            (e.g. if the power varies between program phases but for each phase the power is fixed).
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinOnTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is switched on, it has to remain in this status for at least MinOnTime seconds.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxOnTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is switched on, it must not remain on for more than MaxOnTime seconds.
            This element is for future use and not supported at the moment.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinOffTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is switched off, it has to remain in this status for at least MinOffTime seconds.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxOffTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is switched off, it must not remain off for more than MaxOffTime seconds.
            This element is for future use and not supported at the moment.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="AddEnergySwitchOn" type="xs:double" minOccurs="0" maxOccurs="1" default="0">
        <xs:annotation>
          <xs:documentation>
            Amount of energy in Wh that is additionally needed if a device is switched on. This includes energy required for startup or warmup.
            This element is for future use and not supported at the moment.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="AddCostsSwitchOn" type="xs:double" minOccurs="0" maxOccurs="1" default="0">
        <xs:annotation>
          <xs:documentation>
            Price in Euros for the process of switching on the device. This excludes the cost for the operation. For example, this allows consideration of the limited lifetime of a corresponding relais or other kind of wear caused by switching.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="PowerLevels" type="PowerLevelsType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the power consumption/generation can be controlled but only in a stepwise manner, the available steps are specified in the respective power levels in Watts.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="StandbyPower" type="xs:double" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            The maximum power consumed in standby mode. 
            This is used as threshold to detect inactivity of the device if a timeframe with KeepOnWhileConsumption enabled is used.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="StandbyTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            The time that the power of a device has to remain below StandByPower until it is assumed to be in standby mode.
            This is used to detect inactivity of the device if a timeframe with KeepOnWhileConsumption enabled is used.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="CapabilitiesType">
    <xs:annotation>
      <xs:documentation>
        Encapsulates information about the capabilities of the device
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="CurrentPower" type="CapPowerMeasurementType">
        <xs:annotation>
          <xs:documentation>
            Capability of the device with regard to deriving information about its current power consumption, e.g. measurement or estimation
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Sensor" type="SensorType" minOccurs="0" maxOccurs="10">
        <xs:annotation>
          <xs:documentation>
            Encapsulates information about a sensor attached to the device
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Timestamps" type="CapTimestampType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Specifies if the device is able to deal with absolute timestamps or only with relative timestamps.
            Devices that do not have a synchronized clock (with time server protocols like NTP or radio control like DCF77) or do not have a reliable absolute time source should use relative timestamps.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Interruptions" type="CapInterruptionsType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Specifies if a run of the device can be interrupted or not.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Requests" type="CapRequestsType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Specifies options related to planning requests.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="PowerLevelsType">
    <xs:sequence>
      <xs:element name="PowerLevel" type="xs:double" minOccurs="1" maxOccurs="20">
        <xs:annotation>
          <xs:documentation>
            Power level in Watts [-100000, 100000] positive for consumption, negative for generation.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:simpleType name="SensorTypeRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="Temperature"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="UnitOfMeasurementRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="DegreeCelsius"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:complexType name="SensorType">
    <xs:annotation>
      <xs:documentation>
        Encapsulates information about a sensor attached to the device
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="SensorId" type="xs:string">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the sensor attached to the device. (64 bit identifier in the format XXXX-XXXXXXXX-XX-XX)
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="SensorType" type="xs:string">
        <!-- ref=SensorTypeRefType -->
        <xs:annotation>
          <xs:documentation>
            Type of the sensor. See SensorTypeRefType for known values (e.g. "Temperature").
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="SensorSemantic" type="xs:string" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            String providing information about the semantic of the sensor, e.g. temperature of the water supply
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="UnitOfMeasurement" type="xs:string" minOccurs="1" maxOccurs="1">
        <!-- ref=UnitOfMeasurementRefType -->
        <xs:annotation>
          <xs:documentation>
            String providing information about the unit of measurement.
            See UnitOfMeasurementRefType for known values (e.g. "DegreeCelsius").
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:simpleType name="PowerMeasurementMethodRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="Measurement"/>
      <xs:enumeration value="Estimation"/>
      <xs:enumeration value="None"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:complexType name="CapPowerMeasurementType">
    <xs:annotation>
      <xs:documentation>
        Capability of the device with regard to deriving information about its current power consumption, e.g. measurement or estimation
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Method" type="xs:string">
        <!-- ref=PowerMeasurementMethodRefType -->
        <xs:annotation>
          <xs:documentation>
            String that defines the method of deriving information of the current power consumption.
            A device with a (built-in) power meter should use "Measurement". If it estimates the power consumption by look-up tables or other mechanisms select "Estimation".
            If it is not able to determine the power at all, "None" should be used.
            <p/>
            This information provides a hint about the quality of the power values provided in the device status section. It can be used to determine whether the power data can be used for learning device profile or to decide if it is suitable to be displayed.
            <p/>
            See PowerMeasurementMethodRefType for known values.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="CapTimestampType">
    <xs:annotation>
      <xs:documentation>
        Specifies if the device is able to deal with absolute timestamps or only with relative timestamps
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="AbsoluteTimestamps" type="xs:boolean" default="false">
        <xs:annotation>
          <xs:documentation>
            Bool that indicates if the device is able to deal with absolute timestamps or only with relative timestamps
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="CapInterruptionsType">
    <xs:annotation>
      <xs:documentation>
        Specifies if a run of the device can be interrupted or not.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="InterruptionsAllowed" type="xs:boolean" default="false">
        <xs:annotation>
          <xs:documentation>
            Bool that indicates if a run of the device can be interrupted (paused) or not.
            <p/>
            Allowing interruptions helps the EM to manage energy more flexibly. For instance the EM can interrupt a device in case of unpredictable bad weather conditions or when the user switches on a device with conflicting energy needs and restart it afterwards.
            <p/>
            Not all devices are able to handle interruptions during their runtime. In this case, the device will only be started by the EM and run until it is done.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="CapRequestsType">
    <xs:annotation>
      <xs:documentation>
        Specifies options related to planning requests.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="OptionalEnergy" type="xs:boolean" default="false" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            This capability must be set to "true", if the device uses timeframes in planning requests with minRunningTime!=maxRunningTime or minEnergy!=maxEnergy. This is the case if the device is capable of consuming more (not essentially needed) energy if energy is cheap (e.g. by storing it). Otherwise it should be set to "false".
            <p/>
            As optional energy needs are defined by planning requests it is not possible for the EM to auto-detect this capability until the first optional demand was requested by the device.
            <p/>
            The information whether optional energy is requested is helpful as an EM might provide additional configuration options in a GUI when a device supports consuming optional energy. This might for example include constraints that define under which circumstances (price or ecological limits) optional energy is assigned to the device. Defining that a device is not capable of using optional energy helps the GUI to hide the complexity of defining these constraints from the user.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>

  <xs:simpleType name="StatusRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="On"/>
      <xs:enumeration value="Off"/>
      <xs:enumeration value="Offline"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:complexType name="DeviceStatusType">
    <xs:annotation>
      <xs:documentation>
        A DeviceStatus encapsulates the status information of a device, i.e. all measurements and properties representing the current status of the device
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeviceId" type="DeviceIdType">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="EMSignalsAccepted" type="xs:boolean">
        <xs:annotation>
          <xs:documentation>
            Bool that indicates if the device is currently considering the control signals or recommendations provided by the energy manager or if it is
            in a mode which ignores the signals or recommendations
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Status" type="xs:string">
        <!-- ref=StatusRefType -->
        <xs:annotation>
          <xs:documentation>
            String that provides information of the current status of the device (Offline, On, Off).
            See StatusRefType for known values.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="ErrorCode" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Identifies the current error state of the device. If the code is 0, no error is pending.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="PowerConsumption" type="PowerConsumptionType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Information about the current power consumption of the device, or the power consumption of the device since the last communication.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="SensorValues" type="SensorValuesType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Element encapsulating the measurements of the sensors associated with the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="PowerConsumptionType">
    <xs:annotation>
      <xs:documentation>
        Information about the current power consumption of the device, or the power consumption of the device since the last communication.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="PowerInfo" type="PowerInfoType" minOccurs="1" maxOccurs="10">
        <xs:annotation>
          <xs:documentation>
            Information about a power consumption.
            </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="PowerInfoType">
    <xs:annotation>
      <xs:documentation>
        Information about a power consumption.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="AveragePower" type="xs:int">
        <xs:annotation>
          <xs:documentation>
            Real average power within the interval in Watts.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinPower" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Minimum power value within the interval in Watts.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxPower" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Maximum power within the interval in Watts.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Timestamp" type="RelOrAbsTimeType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Timestamp that represents the end of the averaging interval.
            Although this element is marked as optional it is mandatory in PowerConsumption:PowerInfo.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="AveragingInterval" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Length of the averaging interval in seconds.
            Although this element is marked as optional it is mandatory in PowerConsumption:PowerInfo.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="SensorValuesType">
    <xs:annotation>
      <xs:documentation>
        Provides a sequence of sensor readings
      </xs:documentation>
    </xs:annotation>    
    <xs:sequence>
      <xs:element name="SensorValue" type="SensorValueType" minOccurs="1" maxOccurs="10">
        <xs:annotation>
          <xs:documentation>
            Sequence of sensor measurements
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:complexType name="SensorValueType">
    <xs:annotation>
      <xs:documentation>
        Encapsulates the information about a sensor measurement in a certain time interval
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="SensorId" type="xs:string">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the sensor attached to the device. (64 bit identifier in the format XXXX-XXXXXXXX-XX-XX)
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="AverageValue" type="xs:double">
        <xs:annotation>
          <xs:documentation>
            Average sensor value in the interval
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Timestamp" type="RelOrAbsTimeType">
        <xs:annotation>
          <xs:documentation>
            Timestamp that represents the end of the interval.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="AveragingInterval" type="xs:int">
        <xs:annotation>
          <xs:documentation>
            Length of the averaging interval in seconds
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>

  <xs:complexType name="PlanningRequestType">
    <xs:annotation>
      <xs:documentation>
        A PlanningRequest allows specification of the needs of the device with regard to energy, running time or a certain behaviour of sensor values within certain time frames.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Timeframe" type="TimeframeType" minOccurs="1" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>
            Sequence of timeframes that constitute the planning request.
            Timeframes for one device must not overlap (in terms of earliestStart and latestEnd) as a device can only
            run once at a time.
            In case timeframes overlap the EM has to modify them (e.g. merge or strip) which might lead to unexpected
            behavior.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
  <xs:simpleType name="PreferenceIndifferentAreasRefType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="NoPreference"/>
      <xs:enumeration value="Early"/>
      <xs:enumeration value="Late"/>
    </xs:restriction>
  </xs:simpleType>  
  <xs:complexType name="TimeframeType">
    <xs:annotation>
      <xs:documentation>
        Timeframe as part of a planning request, that allows specification of the energy needs of a device.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeviceId" type="DeviceIdType">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="EarliestStart" type="RelOrAbsTimeType">
        <xs:annotation>
          <xs:documentation>
            Represents the earliest possible time the device can be switched on by the EM.
            The combination of EarliestStart and LatestEnd specifies the interval in which the requested runtime or energy has to be allocated by the EM.
            
            If Min/MaxEnergy is used instead of Min/MaxRuntime, EarliestStart must not be in the past, i.e. contain a negative value, if relative timestamps are used.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="LatestEnd" type="RelOrAbsTimeType">
        <xs:annotation>
          <xs:documentation>
            Represents the latest possible end time the requested minimum runtime (MinRunningTime) or energy (MinEnergy) must be allocated to the device. This means at the given time the device operation must be finished. If a runtime was requested, the latest possible start of operation is LatestEnd-MinRunningTime. In case of an energy request, the latest start depends on the minimum amount of energy (MinEnergy) requested.

            The combination of EarliestStart and LatestEnd specifies the interval in which the requested runtime or energy has to be allocated by the EM.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinRunningTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Minimum running time within the timeframe in seconds.
            If MinRunningTime is 0, the operation of the device in this timeframe is optional.
            Defaults to 0 if MaxRunningTime is set.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxRunningTime" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Maximum running time within the timeframe in seconds.
            If MinRunningTime equals MaxRunningTime, all of the given runtime is required. 
            If MinRunningTime is lower than MaxRunningTime, the amount of runtime given by MinRunningTime is required. The 
            runtime difference between MinRunningTime and MaxRunningTime is optional. That means that the EM will only assign the optional
            runtime to the device if certain conditions like ecological constraints and/or price of energy are met.
            Defaults to MinRunningTime if MinRunningTime is set.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinEnergy" type="xs:double" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Minimum amount of electrical energy in Wh required within the timeframe.
            If MinEnergy is 0, the operation of the device in this timeframe is optional.
            Defaults to 0 if MaxEnergy is set.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxEnergy" type="xs:double" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Maximum amount of electrical energy in Wh that can be consumed or stored within the timeframe.
            If MinEnergy equals MaxEnergy, all of the given amount of energy is required. 
            If MinEnergy is lower than MaxEnergy, the amount of energy given by MinEnergy is required. The 
            energy difference between MinEnergy and MaxEnergy is optional. That means that the EM will only assign the optional
            energy to the device if certain conditions like ecological constraints and/or price of energy are met.
            Defaults to MinEnergy if MinEnergy is set.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MaxPowerConsumption" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is controllable with regard to power consumption, the recommendation of the energy management system will never exceed this value.
            This value overwrites the value provided by the MaxPowerConsumption element in the DeviceInfo. It must be less or equal than the value provided by the DeviceInfo.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="MinPowerConsumption" type="xs:int" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            If the device is controllable with regard to power consumption, the recommendation of the energy management system will never go below this value.
            This value overwrites the value provided by the MinPowerConsumption element in the DeviceInfo. It must be greater or equal than the value provided by the DeviceInfo.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="KeepOnWhileConsumption" type="xs:boolean" minOccurs="0" maxOccurs="1" default="false">
        <xs:annotation>
          <xs:documentation>
            Flag that indicates if the energy manager has to keep the control signal on, until no power consumption can be observed anymore.
            The DeviceInfo StandbyPower and StandbyTime parameters are used to detect inactivity. 
            Default values are used for these parameters if not specified by the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="PreferenceIndifferentAreas" type="xs:string" minOccurs="0" maxOccurs="1" default="NoPreference">
        <!-- ref=PreferenceIndifferentAreasRefType -->
        <xs:annotation>
          <xs:documentation>
            Indicates the preference within areas of indifferent price.
            That means if there are multiple possibilities to schedule a device with only small difference in prices
            the one according to the this preference is selected.
            If "Early" is selected, the earliest possibility is selected, for "Late" the latest. The default "NoPreference" 
            does not specify a preference, so the decision is made by the EM.
            See PreferenceIndifferentAreasRefType for known values.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>

  <xs:complexType name="DeviceControlType">
    <xs:annotation>
      <xs:documentation>
        Contains operation and power consumption recommendations for a device.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeviceId" type="DeviceIdType">
        <xs:annotation>
          <xs:documentation>
            Unique identification of the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="On" type="xs:boolean">
        <xs:annotation>
          <xs:documentation>
            Switch on/off recommendation.
            If On=true, the EM recommends that the device should be switched on or run its program.
            For interruptible devices On=false is a switch-off recommendation indicating that the device should switch off or pause.
            A device should follow the recommendation as soon as possible. Otherwise the device might interfere with other devices 
            managed by the EM.
            Note that the device should only accept a recommendation if stable operation is guaranteed (no risk of damage or safety issues).
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="RecommendedPowerConsumption" type="xs:double" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>
            Recommended power consumption in Watts
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="Timestamp" type="RelOrAbsTimeType">
        <xs:annotation>
          <xs:documentation>
            Timestamp of the generation of the message. 
            Note that this does not determine the activation time of the switch-on or power recommendation. Recommendations are supposed to be applied immediately by the device.
          </xs:documentation>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
  </xs:complexType>
</xs:schema>
"""



"""
abstract base class on which all device implementations are based
"""

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List

from .sensor import Sensor, Sensors


@dataclass
class DiscoveryInformation:
    tested_endpoints: str = ""
    status: str = ""
    access: str = ""
    exception: Exception | None = None
    remark: str = ""
    device: str = ""


@dataclass
class DeviceInformation:
    id: str
    serial: str
    name: str
    type: str
    manufacturer: str
    sw_version: str
    additional: dict[str, str | int] = field(default_factory=dict)
    measurementsCount: int | None = None
    parameterCount: int | None = None

    def asDict(self) -> dict[str, Any]:
        """Returns the values as a dict"""
        return asdict(self)

    def __str__(self):
        return f"{self.serial} {self.name}"


class Device(ABC):
    """abstract base class on which all device implementations are based."""

    @abstractmethod
    async def get_sensors(self, deviceID: str | None = None) -> Sensors:
        """Returns a list of all supported sensors"""

    @abstractmethod
    async def new_session(self) -> bool:
        """Starts a new session"""

    @abstractmethod
    async def device_info(self) -> dict:
        """Return a Dict with basic device information"""

    @abstractmethod
    async def device_list(self) -> dict[str, DeviceInformation]:
        """List of all devices"""

    @abstractmethod
    async def read(self, sensors: Sensors, deviceID: str | None = None) -> bool:
        """Updates all sensors"""

    @abstractmethod
    async def close_session(self) -> None:
        """Closes the session"""

    @abstractmethod
    async def detect(self, ip: str) -> List[DiscoveryInformation]:
        """Try to detect SMA devices"""

    @abstractmethod
    async def get_debug(self) -> Dict[str, Any]:
        """Return a dict with all debug information."""

    def set_options(self, options: Dict[str, Any]) -> None:
        """Set options"""
        pass

    async def set_parameter(
        self, sensor: Sensor, value: int, deviceID: str | None = None
    ) -> None:
        """Set Parameters."""
        pass



""" Speedwire Discovery  """

import asyncio
import logging
import socket
import struct

from .definitions_speedwire import speedwireHeader

_LOGGER = logging.getLogger(__name__)


class Discovery:
    """Class for the detection of SMA Devices in the local network."""

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        """init"""
        self.loop = loop
        self.transport: asyncio.BaseTransport
        self.addr = "239.12.255.254"
        self.port = 9522
        self.discovered: list[tuple[str, int]] = []

    def getDiscoverySocket(self) -> socket.socket:
        addrinfo = socket.getaddrinfo(self.addr, None)[0]
        sock = socket.socket(addrinfo[0], socket.SOCK_DGRAM)
        sock.setsockopt(
            socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, struct.pack("@i", 1)
        )
        return sock

    async def run(self) -> list:
        """Start the Task"""
        sock = self.getDiscoverySocket()
        self.loop.create_future()
        await self.loop.create_datagram_endpoint(
            lambda: self,  # type: ignore[type-var]
            sock=sock,
        )
        for i in range(0, 3):
            await asyncio.sleep(0.5)
            self.sendDiscoveryRequest()
        await asyncio.sleep(0.5)
        return self.discovered

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = transport
        self.sendDiscoveryRequest()

    def sendDiscoveryRequest(self) -> None:
        """Send a discovery Request"""
        _LOGGER.warn("Sending Discovery Request")
        self.transport.sendto(  # type: ignore[attr-defined]
            bytes.fromhex("534d4100000402a0ffffffff0000002000000000"),
            (self.addr, self.port),
        )

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """Datagram received"""
        msg = speedwireHeader.from_packed(data[0:18])
        if not msg.isDiscoveryResponse():
            _LOGGER.warning("Ignoring %s", msg)
            return
        if addr not in self.discovered:
            self.discovered.append(addr)

    def error_received(self, exc: Exception) -> None:
        """Called by error."""
        _LOGGER.error("%s error occurred: %s", type(exc), exc)

    def connection_lost(self, exc: Exception) -> None:
        """Called by connection lost."""
        _LOGGER.error("Socket closed, stop the event loop %s %s", type(exc), exc)



import asyncio
import random
import uuid
from datetime import datetime
from typing import Awaitable, Callable, Union
from zoneinfo import ZoneInfo

from pysmaplus.semp.const import callbackAction

from .device import sempDevice
from .SEMPhttpd import SEMPhttpServer
from .ssdp import async_create_upnp_datagram_endpoint


class semp:
    def __init__(
        self,
        ip: str,
        port: int,
        timezone: ZoneInfo | None = None,
        callback: Callable[[callbackAction], Awaitable[None]] | None = None,
    ):
        self.ip = ip
        self.port = port
        self.uuid = self.getUUID()
        self.http = SEMPhttpServer(self.ip, self.port, self.uuid, timezone, callback)

    def getUUID(self) -> str:
        """Generate a UUID based on the IP and the used port"""
        rnd = random.Random()
        rnd.seed(self.ip + str(self.port))
        return str(uuid.UUID(int=rnd.getrandbits(128), version=4))

    async def start(self, embeddedHttpd: bool = True):
        if embeddedHttpd:
            await self.http.startWebserver()
        await asyncio.gather(asyncio.to_thread(self.http.loadSempSchema))
        await async_create_upnp_datagram_endpoint(
            self.ip, True, self.ip, self.port, self.uuid
        )

    def addDevice(self, dev: sempDevice):
        self.http.addDevice(dev)

    def removeDevice(self, dev: Union[str | sempDevice]):
        if isinstance(dev, sempDevice):
            self.http.removeDevice(dev.deviceId)
        else:
            self.http.removeDevice(dev)

    def getDevice(self, deviceId: Union[str | sempDevice]) -> sempDevice:
        if isinstance(deviceId, sempDevice):
            return self.getDevice(deviceId.deviceId)
        return self.http.getDevice(deviceId)

    def getStatus(self) -> str:
        if len(self.http.history) == 0:
            return "Not Connected"
        r = self.http.history[-1]
        diff = datetime.now().timestamp() - r.timemsec
        if diff < 130:
            return "Connected"
        return "Connection lost"

    def getRoutes(self) -> list[dict]:
        return [
            {"path": "/sempinfo", "callback": {"GET": self.http.getStatusPage}},
            {
                "path": "/uuid:" + self.uuid + "/description.xml",
                "callback": {"GET": self.http.getUUIDPage},
            },
            {
                "path": "/semp/",
                "callback": {"GET": self.http.getSemp, "POST": self.http.postSemp},
            },
        ]
